from typing import List, Optional

from caluma.caluma_form.models import Form
from caluma.caluma_user.models import BaseUser
from caluma.utils import update_model

from . import domain_logic, models


def start_case(
    workflow: models.Workflow,
    user: BaseUser,
    form: Optional[Form] = None,
    parent_work_item: Optional[models.WorkItem] = None,
    context: Optional[dict] = None,
    **kwargs
) -> models.Case:
    """
    Start a case of a given workflow (just like `saveCase`).

    Similar to Case.objects.create(), but with the same business logic as
    used in the `saveCase` mutation.

    >>> start_case(
    ...     workflow=Workflow.objects.get(pk="my-wf"),
    ...     form=Form.objects.get(pk="my-form")),
    ...     user=AnonymousUser()
    ... )
    <Case: Case object (some-uuid)>
    """
    data = {"workflow": workflow, "form": form, "parent_work_item": parent_work_item}

    data.update(kwargs)

    validated_data = domain_logic.StartCaseLogic.pre_start(
        domain_logic.StartCaseLogic.validate_for_start(data), user
    )

    case = models.Case.objects.create(**validated_data)

    return domain_logic.StartCaseLogic.post_start(case, user, parent_work_item, context)


def complete_work_item(
    work_item: models.WorkItem, user: BaseUser, context: Optional[dict] = None
) -> models.WorkItem:
    """
    Complete a work item (just like `completeWorkItem`).

    >>> complete_work_item(
    ...     work_item=models.WorkItem.first(),
    ...     user=AnonymousUser()
    ... )
    <WorkItem: WorkItem object (some-uuid)>
    ```
    """
    domain_logic.CompleteWorkItemLogic.validate_for_complete(work_item, user)
    validated_data = domain_logic.CompleteWorkItemLogic.pre_complete(
        work_item, {}, user, context
    )

    update_model(work_item, validated_data)

    domain_logic.CompleteWorkItemLogic.post_complete(work_item, user, context)

    return work_item


def skip_work_item(
    work_item: models.WorkItem, user: BaseUser, context: Optional[dict] = None
) -> models.WorkItem:
    """
    Skip a work item (just like `SkipWorkItem`).

    >>> skip_work_item(
    ...     work_item=models.WorkItem.first(),
    ...     user=AnonymousUser()
    ... )
    <WorkItem: WorkItem object (some-uuid)>
    """
    domain_logic.SkipWorkItemLogic.validate_for_skip(work_item)

    validated_data = domain_logic.SkipWorkItemLogic.pre_skip(
        work_item, {}, user, context
    )

    update_model(work_item, validated_data)

    domain_logic.SkipWorkItemLogic.post_skip(work_item, user, context)

    return work_item


def cancel_case(
    case: models.Case, user: BaseUser, context: Optional[dict] = None
) -> models.Case:
    """
    Cancel a case and its pending work items (just like `CancelCase`).

    >>> cancel_case(
    ...     case=models.Case.first(),
    ...     user=AnonymousUser()
    ... )
    <Case: Case object (some-uuid)>
    """
    domain_logic.CancelCaseLogic.validate_for_cancel(case)

    validated_data = domain_logic.CancelCaseLogic.pre_cancel(case, {}, user, context)

    update_model(case, validated_data)

    domain_logic.CancelCaseLogic.post_cancel(case, user, context)

    return case


def cancel_work_item(
    work_item: models.WorkItem, user: BaseUser, context: Optional[dict] = None
) -> models.WorkItem:
    """
    Cancel a work item (just like `CancelWorkItem`).

    >>> cancel_work_item(
    ...     work_item=models.WorkItem.first(),
    ...     user=AnonymousUser()
    ... )
    <WorkItem: WorkItem object (some-uuid)>
    """
    domain_logic.CancelWorkItemLogic.validate_for_cancel(work_item)

    validated_data = domain_logic.CancelWorkItemLogic.pre_cancel(
        work_item, {}, user, context
    )

    update_model(work_item, validated_data)

    domain_logic.CancelWorkItemLogic.post_cancel(work_item, user, context)

    return work_item


def suspend_case(
    case: models.Case, user: BaseUser, context: Optional[dict] = None
) -> models.Case:
    """
    Suspend a case (just like `SuspendCase`).

    >>> suspend_case(
    ...     case=models.Case.first(),
    ...     user=AnonymousUser()
    ... )
    <Case: Case object (some-uuid)>
    """
    domain_logic.SuspendCaseLogic.validate_for_suspend(case)

    validated_data = domain_logic.SuspendCaseLogic.pre_suspend(case, {}, user, context)

    update_model(case, validated_data)

    domain_logic.SuspendCaseLogic.post_suspend(case, user, context)

    return case


def suspend_work_item(
    work_item: models.WorkItem, user: BaseUser, context: Optional[dict] = None
) -> models.WorkItem:
    """
    Suspend a work item (just like `SuspendWorkItem`).

    >>> suspend_work_item(
    ...     work_item=models.WorkItem.first(),
    ...     user=AnonymousUser()
    ... )
    <WorkItem: WorkItem object (some-uuid)>
    """
    domain_logic.SuspendWorkItemLogic.validate_for_suspend(work_item)

    validated_data = domain_logic.SuspendWorkItemLogic.pre_suspend(
        work_item, {}, user, context
    )

    update_model(work_item, validated_data)

    domain_logic.SuspendWorkItemLogic.post_suspend(work_item, user, context)

    return work_item


def resume_case(
    case: models.Case, user: BaseUser, context: Optional[dict] = None
) -> models.Case:
    """
    Resume a case (just like `ResumeCase`).

    >>> resume_case(
    ...     case=models.Case.first(),
    ...     user=AnonymousUser()
    ... )
    <Case: Case object (some-uuid)>
    """
    domain_logic.ResumeCaseLogic.validate_for_resume(case)

    validated_data = domain_logic.ResumeCaseLogic.pre_resume(case, {}, user, context)

    update_model(case, validated_data)

    domain_logic.ResumeCaseLogic.post_resume(case, user, context)

    return case


def reopen_case(
    case: models.Case,
    work_items: List[models.WorkItem],
    user: BaseUser,
    context: Optional[dict] = None,
) -> models.Case:
    """
    Reopen a case (just like `ReopenCase`).

    >>> reopen_case(
    ...     case=models.Case.first(),
    ...     work_items=[...]
    ...     user=AnonymousUser()
    ... )
    <Case: Case object (some-uuid)>
    """

    domain_logic.ReopenCaseLogic.validate_for_reopen(case, work_items)

    domain_logic.ReopenCaseLogic.pre_reopen(case, work_items, user, context)

    domain_logic.ReopenCaseLogic.do_reopen(case, work_items)

    domain_logic.ReopenCaseLogic.post_reopen(case, work_items, user, context)

    return case


def resume_work_item(
    work_item: models.WorkItem, user: BaseUser, context: Optional[dict] = None
) -> models.WorkItem:
    """
    Resume a work item (just like `ResumeWorkItem`).

    >>> resume_work_item(
    ...     work_item=models.WorkItem.first(),
    ...     user=AnonymousUser()
    ... )
    <WorkItem: WorkItem object (some-uuid)>
    """
    domain_logic.ResumeWorkItemLogic.validate_for_resume(work_item)

    validated_data = domain_logic.ResumeWorkItemLogic.pre_resume(
        work_item, {}, user, context
    )

    update_model(work_item, validated_data)

    domain_logic.ResumeWorkItemLogic.post_resume(work_item, user, context)

    return work_item


def redo_work_item(
    work_item: models.WorkItem, user: BaseUser, context: Optional[dict] = None
) -> models.WorkItem:
    """
    Redo a work item.

    >>> redo_work_item(
    ...     work_item=models.WorkItem.objects.first(),
    ...     user=AnonymousUser()
    ... )
    <WorkItem: WorkItem object (some-uuid)>
    """
    domain_logic.RedoWorkItemLogic.validate_for_redo(work_item)

    validated_data = domain_logic.RedoWorkItemLogic.pre_redo(
        work_item, {}, user, context
    )

    domain_logic.RedoWorkItemLogic.set_succeeding_work_item_status_redo(work_item)

    update_model(work_item, validated_data)

    domain_logic.RedoWorkItemLogic.post_redo(work_item, user, context)

    return work_item



from collections.abc import Iterable

import graphene
from django.core.exceptions import ImproperlyConfigured
from django.db.models.query import QuerySet
from django.db.models.query_utils import DeferredAttribute
from graphene.relay import PageInfo
from graphene.relay.connection import ConnectionField
from graphene_django import types
from graphene_django.fields import DjangoConnectionField
from graphene_django.rest_framework import serializer_converter
from graphene_django.utils import maybe_queryset
from graphql_relay import cursor_to_offset, get_offset_with_default, offset_to_cursor

from .pagination import connection_from_array, connection_from_array_slice


def enum_type_from_field(
    name, field=None, choices=None, description=None, serializer_field=None
):
    """Create enum type for a given django model field.

    The model field must be of the right type (CharField) and
    have it's choices configured properly for this to work.

    Alternatively, you may pass a `choices` list of singular
    values (not CharField choices tuples!) as choices to use.

    Usually, you will want to pass the `serializer_field` parameter
    as well to register the type, so it will be correctly used in
    the appropriate places.

    Examples:
    >>> Field = enum_type_from_field("Field", field=MyModel.some_choice_field)
    Field

    >>> Field2 = enum_type_from_field("Field2", choices=["foo", "bar"])
    Field2

    """
    if field and isinstance(field, DeferredAttribute):
        field = field.field  # pragma: no cover
        choices = [(key.upper(), key) for key, _ in field.choices]
    else:
        # Force choices to be always uppercase
        choices = [(key.upper(), key) for key in choices]

    the_type = graphene.Enum(name, choices, description=description)

    if serializer_field:
        serializer_converter.get_graphene_type_from_serializer_field.register(
            serializer_field, lambda field: the_type
        )

    return the_type


class Node(object):
    """Base class to define queryset filters for all nodes."""

    # will be set in caluma_core.AppConfig.ready hook, see apps.py
    # to avoid recursive import error
    visibility_classes = None

    @classmethod
    def get_queryset(cls, queryset, info):
        if cls.visibility_classes is None:
            raise ImproperlyConfigured(
                "check that app `caluma.caluma_core` is part of your `INSTALLED_APPS` "
                "or custom node has `visibility_classes` properly assigned."
            )

        for visibility_class in cls.visibility_classes:
            queryset = visibility_class().filter_queryset(cls, queryset, info)

        return queryset.select_related()


class DjangoObjectType(Node, types.DjangoObjectType):
    """Django object type implementing default get_queryset with visibility layer."""

    class Meta:
        abstract = True


class CountableConnectionBase(graphene.Connection):
    """Connection subclass that supports totalCount."""

    class Meta:
        abstract = True

    total_count = graphene.Int()

    def resolve_total_count(self, info, **kwargs):
        try:
            # DjangoConnectionField sets the length already
            return self.length
        except AttributeError:
            if isinstance(self.iterable, QuerySet):  # pragma: no cover
                return self.iterable.count()
            return len(self.iterable)


class DjangoConnectionField(DjangoConnectionField):
    """
    Custom DjangoConnectionField with fix for hasNextPage/hasPreviousPage.

    This can be removed, when (or better if)
    https://github.com/graphql-python/graphql-relay-py/issues/12
    is resolved.

    WARNING: This is a one to one copy of the method in graphene django with the
    difference that we use our customized connection_from_array_slice function
    and that we only count on the database if pagination is actually used:
    https://github.com/graphql-python/graphene-django/blob/775644b5369bdc5fbb45d3535ae391a069ebf9d4/graphene_django/fields.py#L136
    """

    @classmethod
    def resolve_connection(cls, connection, args, iterable, max_limit=None):
        # Remove the offset parameter and convert it to an after cursor.
        offset = args.pop("offset", None)
        after = args.get("after")
        if offset:
            if after:
                offset += cursor_to_offset(after) + 1
            # input offset starts at 1 while the graphene offset starts at 0
            args["after"] = offset_to_cursor(offset - 1)

        iterable = maybe_queryset(iterable)

        if isinstance(iterable, QuerySet):
            # only query count on database when pagination is needed
            # resolve_connection may be removed again once following issue is fixed:
            # https://github.com/graphql-python/graphene-django/issues/177
            if all(
                args.get(pagination_arg) is None
                for pagination_arg in ["before", "after", "first", "last"]
            ):
                list_length = len(iterable)
            else:
                list_length = iterable.count()
        else:  # pragma: no cover
            list_length = len(iterable)
        list_slice_length = (
            min(max_limit, list_length) if max_limit is not None else list_length
        )

        # If after is higher than list_length, connection_from_array_slice
        # would try to do a negative slicing which makes django throw an
        # AssertionError
        after = min(get_offset_with_default(args.get("after"), -1) + 1, list_length)

        # This is commented out because we disable this functionality in our
        # settings explicitly.
        # if max_limit is not None and args.get("first", None) is None:
        #     if args.get("last", None) is not None:
        #         after = list_length - args["last"]
        #     else:
        #         args["first"] = max_limit

        connection = connection_from_array_slice(
            iterable[after:],
            args,
            slice_start=after,
            array_length=list_length,
            array_slice_length=list_slice_length,
            connection_type=connection,
            edge_type=connection.Edge,
            page_info_type=PageInfo,
        )
        connection.iterable = iterable
        connection.length = list_length
        return connection


class ConnectionField(ConnectionField):
    """
    Custom ConnectionField with fix for hasNextPage/hasPreviousPage.

    This can be removed, when (or better if)
    https://github.com/graphql-python/graphql-relay-py/issues/12
    is resolved.

    WARNING: This is a one to one copy of the method in graphene with the
    difference that we use our customized connection_from_array function:
    https://github.com/graphql-python/graphene/blob/61f0d8a8e09086fc74ad51d9ec80004674bd91f1/graphene/relay/connection.py#L146
    """

    @classmethod
    def resolve_connection(cls, connection_type, args, resolved):
        if isinstance(resolved, connection_type):  # pragma: no cover
            return resolved

        assert isinstance(resolved, Iterable), (
            f"Resolved value from the connection field has to be an iterable or instance of {connection_type}. "
            f'Received "{resolved}"'
        )
        connection = connection_from_array(
            resolved,
            args,
            connection_type=connection_type,
            edge_type=connection_type.Edge,
            page_info_type=PageInfo,
        )
        connection.iterable = resolved
        return connection



from functools import reduce

import graphene
from django.core import exceptions
from django.db import ProgrammingError
from django.db.models import Q
from django.forms import BooleanField
from django.utils import translation
from django_filters.constants import EMPTY_VALUES
from django_filters.rest_framework import Filter, FilterSet, MultipleChoiceFilter
from graphene import Enum, InputObjectType, List
from graphene_django.forms.converter import convert_form_field
from graphene_django.registry import get_global_registry

from ..caluma_core.filters import (
    CompositeFieldClass,
    GlobalIDFilter,
    GlobalIDMultipleChoiceFilter,
    MetaFilterSet,
    SearchFilter,
)
from ..caluma_core.forms import GlobalIDFormField
from ..caluma_core.ordering import AttributeOrderingFactory, MetaFieldOrdering
from ..caluma_core.relay import extract_global_id
from ..caluma_form.models import Answer, DynamicOption, Form, Question
from ..caluma_form.ordering import AnswerValueOrdering
from . import models, validators


class FormFilterSet(MetaFilterSet):
    search = SearchFilter(fields=("slug", "name", "description"))
    slugs = MultipleChoiceFilter(field_name="slug")
    questions = MultipleChoiceFilter(field_name="questions__slug")

    class Meta:
        model = models.Form
        fields = (
            "name",
            "description",
            "is_published",
            "is_archived",
            "questions",
        )


class FormOrderSet(FilterSet):
    meta = MetaFieldOrdering()
    attribute = AttributeOrderingFactory(
        models.Form,
        fields=[
            "created_at",
            "modified_at",
            "slug",
            "name",
            "description",
            "is_published",
            "is_archived",
        ],
    )

    class Meta:
        model = models.Form
        fields = ("meta", "attribute")


class OptionFilterSet(MetaFilterSet):
    search = SearchFilter(fields=("slug", "label"))

    class Meta:
        model = models.Option
        fields = ("slug", "label", "is_archived")


class OptionOrderSet(FilterSet):
    meta = MetaFieldOrdering()
    attribute = AttributeOrderingFactory(
        models.Option,
        fields=["created_at", "modified_at", "slug", "label", "is_archived"],
    )

    class Meta:
        model = models.Question
        fields = ("meta", "attribute")


class QuestionOrderSet(FilterSet):
    meta = MetaFieldOrdering()
    attribute = AttributeOrderingFactory(
        models.Question,
        fields=[
            "created_at",
            "modified_at",
            "slug",
            "label",
            "type",
            "is_required",
            "is_hidden",
            "is_archived",
            "placeholder",
            "info_text",
            "hint_text",
            "calc_expression",
        ],
    )

    class Meta:
        model = models.Question
        fields = ("meta", "attribute")


class AnswerLookupMode(Enum):
    EXACT = "exact"
    STARTSWITH = "startswith"
    CONTAINS = "contains"
    ICONTAINS = "icontains"
    INTERSECTS = "intersects"
    ISNULL = "isnull"
    IN = "in"

    GTE = "gte"
    GT = "gt"
    LTE = "lte"
    LT = "lt"


class AnswerHierarchyMode(Enum):
    DIRECT = "DIRECT"
    FAMILY = "FAMILY"


class HasAnswerFilterType(InputObjectType):
    """
    Lookup type to search document structures.

    When using lookup `ISNULL`, the provided `value` will be ignored.
    """

    question = graphene.ID(required=True)
    value = graphene.types.generic.GenericScalar(required=False)
    lookup = AnswerLookupMode()
    hierarchy = AnswerHierarchyMode()


class HasAnswerFilterField(CompositeFieldClass):
    pass


class HasAnswerFilter(Filter):
    field_class = HasAnswerFilterField

    def __init__(self, *args, **kwargs):
        self.document_id = kwargs.pop("document_id")
        super().__init__(*args, **kwargs)

    VALID_LOOKUPS = {
        "text": [
            AnswerLookupMode.EXACT,
            AnswerLookupMode.STARTSWITH,
            AnswerLookupMode.CONTAINS,
            AnswerLookupMode.ICONTAINS,
            AnswerLookupMode.ISNULL,
            AnswerLookupMode.IN,
        ],
        "integer": [
            AnswerLookupMode.EXACT,
            AnswerLookupMode.LT,
            AnswerLookupMode.LTE,
            AnswerLookupMode.GT,
            AnswerLookupMode.GTE,
            AnswerLookupMode.ISNULL,
            AnswerLookupMode.IN,
        ],
        "choice": [
            AnswerLookupMode.EXACT,
            AnswerLookupMode.IN,
            AnswerLookupMode.ISNULL,
        ],
        "multiple_choice": [
            AnswerLookupMode.EXACT,
            AnswerLookupMode.CONTAINS,
            AnswerLookupMode.INTERSECTS,
            AnswerLookupMode.ISNULL,
        ],
    }
    VALID_LOOKUPS["date"] = VALID_LOOKUPS["integer"]
    VALID_LOOKUPS["float"] = VALID_LOOKUPS["integer"]
    VALID_LOOKUPS["dynamic_choice"] = VALID_LOOKUPS["choice"]
    VALID_LOOKUPS["dynamic_multiple_choice"] = VALID_LOOKUPS["multiple_choice"]
    VALID_LOOKUPS["textarea"] = VALID_LOOKUPS["text"]
    VALID_LOOKUPS["datetime"] = VALID_LOOKUPS["integer"]
    VALID_LOOKUPS["calculated_float"] = VALID_LOOKUPS["float"]

    def filter(self, qs, value):
        if value in EMPTY_VALUES:  # pragma: no cover
            return qs

        for expr in value:
            qs = self.apply_expr(qs, expr)
        return qs

    def apply_expr(self, qs, expr):
        lookup = expr.get("lookup", self.lookup_expr)
        lookup_expr = (hasattr(lookup, "value") and lookup.value) or lookup

        question_slug = expr["question"]
        match_value = expr.get("value")
        if lookup == AnswerLookupMode.ISNULL:
            match_value = True

        hierarchy = expr.get("hierarchy", AnswerHierarchyMode.FAMILY)

        question = models.Question.objects.get(slug=question_slug)
        self._validate_lookup(question, lookup)

        answer_value = "value"
        if question.type == models.Question.TYPE_DATE:
            answer_value = "date"

        answers = models.Answer.objects.all()

        if lookup == AnswerLookupMode.INTERSECTS:
            inner_lookup = "exact"
            if question.type in (
                models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
                models.Question.TYPE_MULTIPLE_CHOICE,
            ):
                inner_lookup = "contains"

            exprs = [
                Q(**{f"value__{inner_lookup}": val, "question__slug": question_slug})
                for val in match_value
            ]
            # connect all expressions with OR
            answers = answers.filter(reduce(lambda a, b: a | b, exprs))
        else:
            answers = answers.filter(
                **{
                    f"{answer_value}__{lookup_expr}": match_value,
                    "question__slug": question_slug,
                }
            )

        if hierarchy == AnswerHierarchyMode.FAMILY:
            return qs.filter(
                **{f"{self.document_id}__in": answers.values("document__family")}
            )
        else:
            return qs.filter(**{f"{self.document_id}__in": answers.values("document")})

    def _validate_lookup(self, question, lookup):
        try:
            valid_lookups = self.VALID_LOOKUPS[question.type]
        except KeyError:  # pragma: no cover
            # Not covered in tests - this only happens when you add a new
            # question type and forget to update the lookup config above.  In
            # that case, the fix is simple - go up a few lines and adjust the
            # VALID_LOOKUPS dict.
            raise ProgrammingError(
                f"Valid lookups not configured for question type {question.type}"
            )

        if lookup not in valid_lookups:
            lookup = (hasattr(lookup, "value") and lookup.value) or lookup
            raise exceptions.ValidationError(
                f"Invalid lookup for question slug={question.slug} ({question.type.upper()}): {lookup.upper()}"
            )

    @staticmethod
    @convert_form_field.register(HasAnswerFilterField)
    def convert_meta_value_field(field):
        # the converted type must be list-of-filter, as we need to apply
        # multiple conditions
        converted = List(HasAnswerFilterType)
        get_global_registry().register_converted_field(field, converted)
        return converted


class SearchLookupMode(Enum):
    STARTSWITH = "startswith"
    CONTAINS = "icontains"
    TEXT = "search"


class SearchAnswersFilterType(InputObjectType):
    """
    Lookup type to search in answers.

    You may pass in a list of question slugs and/or a list of form slugs to define
    which answers to search. If you pass in one or more forms, answers to the
    questions in that form will be searched. If you pass in one or more question
    slug, the corresponding answers are searched. If you pass both, a superset
    of both is searched (ie. they do not limit each other).
    """

    questions = List(graphene.ID)
    forms = List(graphene.ID)
    value = graphene.types.generic.GenericScalar(required=True)
    lookup = SearchLookupMode(required=False)


class SearchAnswersFilterField(CompositeFieldClass):
    pass


class SearchAnswersFilter(Filter):
    field_class = SearchAnswersFilterField

    FIELD_MAP = {
        Question.TYPE_TEXT: "value",
        Question.TYPE_TEXTAREA: "value",
        Question.TYPE_DATE: "date",
        Question.TYPE_CHOICE: "value",
        Question.TYPE_MULTIPLE_CHOICE: "value",
        Question.TYPE_DYNAMIC_CHOICE: "value",
        Question.TYPE_DYNAMIC_MULTIPLE_CHOICE: "value",
        Question.TYPE_INTEGER: "value",
        Question.TYPE_FLOAT: "value",
    }

    def __init__(self, *args, **kwargs):
        self.document_id = kwargs.pop("document_id")
        super().__init__(*args, **kwargs)

    def filter(self, qs, value):
        if value in EMPTY_VALUES:  # pragma: no cover
            return qs

        assert isinstance(value, list)

        for val in value:
            if val in EMPTY_VALUES:  # pragma: no cover
                continue
            qs = self._apply_filter(qs, val)
        return qs

    def _get_questions(self, value):
        raw_questions = value.get("questions")
        raw_forms = value.get("forms")
        questions = Question.objects.none()

        if not raw_questions and not raw_forms:
            raise exceptions.ValidationError(
                '"forms" and/or "questions" parameter must be set'
            )

        if raw_questions:
            questions = self._validate_and_get_questions(raw_questions)

        if raw_forms:
            forms = self._validate_and_get_forms(raw_forms)
            form_questions = Form.get_all_questions(forms)
            # Combine querysets: All questions of the given forms, as well as the
            # explicitly-requested questions.
            questions = questions | form_questions.filter(
                type__in=self.FIELD_MAP.keys()
            )

        return questions

    def _apply_filter(self, qs, value):
        questions = self._get_questions(value)

        for word in value["value"].split():
            answers_with_word = self._answers_with_word(
                questions,
                word,
                value.get("lookup", SearchLookupMode.CONTAINS.value),
                value.get("forms"),
            )
            qs = qs.filter(
                **{
                    f"{self.document_id}__in": answers_with_word.values(
                        "document__family"
                    )
                }
            )

        return qs

    def _word_lookup_for_question(self, question, word, lookup):
        if question.type not in (
            Question.TYPE_CHOICE,
            Question.TYPE_DYNAMIC_CHOICE,
            Question.TYPE_MULTIPLE_CHOICE,
            Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
        ):
            return Q(
                **{
                    f"{self.FIELD_MAP[question.type]}__{lookup}": word,
                    "question": question,
                }
            )

        # (Multiple) choice lookups need more specific treatment...
        lang = translation.get_language()
        is_multiple = question.type in (
            Question.TYPE_MULTIPLE_CHOICE,
            Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
        )
        is_dynamic = question.type in (
            Question.TYPE_DYNAMIC_CHOICE,
            Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
        )

        # find all options of our given question that match the
        # word, then use their slugs for lookup
        if is_dynamic:
            matching_options = (
                DynamicOption.objects.all()
                .filter(question=question, **{f"label__{lang}__{lookup}": word})
                .values_list("slug", flat=True)
            )
        else:
            matching_options = question.options.filter(
                **{f"label__{lang}__{lookup}": word}
            ).values_list("slug", flat=True)

        if not matching_options:
            # no labels = no results
            return Q(value=False) & Q(value=True)

        filt = "value__contains" if is_multiple else "value"
        return reduce(
            lambda a, b: a | b, [Q(**{filt: slug}) for slug in matching_options]
        )

    def _answers_with_word(self, questions, word, lookup, form_slugs):
        if not questions:
            return Answer.objects.none()

        exprs = [
            self._word_lookup_for_question(question, word, lookup)
            for question in questions
        ]

        # join expressions with OR
        answer_qs = Answer.objects.filter(reduce(lambda a, b: a | b, exprs))

        # add form filter if given,
        # otherwise it would return all answers of the question filter ignoring the form filter
        if form_slugs not in EMPTY_VALUES:
            answer_qs = answer_qs.filter(document__form__pk__in=form_slugs)

        return answer_qs

    def _validate_and_get_questions(self, question_slugs):
        res = []
        for q_slug in question_slugs:
            question = Question.objects.get(pk=q_slug)
            if question.type not in self.FIELD_MAP:
                raise exceptions.ValidationError(
                    f"Questions of type {question.type} cannot be used in searchAnswers"
                )
            res.append(question)
        return Question.objects.filter(pk__in=res)

    @staticmethod
    def _validate_and_get_forms(form_slugs):
        forms = Form.objects.filter(slug__in=form_slugs)
        not_found = [
            x for x in form_slugs if x not in forms.values_list("slug", flat=True)
        ]
        if not_found:
            not_found_string = ", ".join(not_found)
            raise exceptions.ValidationError(
                f"Following forms could not be found: {not_found_string}"
            )
        return form_slugs

    @staticmethod
    @convert_form_field.register(SearchAnswersFilterField)
    def convert_meta_value_field(field):
        registry = get_global_registry()
        converted = registry.get_converted_field(field)
        if converted:
            # TODO: can this be removed, as it's never the case
            # anymore with graphene 3.0?
            # Doesn't seem to have ill effects...
            return converted  # pragma: no cover

        converted = List(SearchAnswersFilterType)
        registry.register_converted_field(field, converted)
        return converted


class DocumentFilterSet(MetaFilterSet):
    id = GlobalIDFilter()
    search = SearchFilter(
        fields=(
            "form__slug",
            "form__name",
            "form__description",
            "answers__value",
            "answers__files__name",
        )
    )
    root_document = GlobalIDFilter(field_name="family")
    forms = GlobalIDMultipleChoiceFilter(field_name="form")

    has_answer = HasAnswerFilter(document_id="pk")
    search_answers = SearchAnswersFilter(document_id="pk")

    class Meta:
        model = models.Document
        fields = ("form", "forms", "search", "id")


class DocumentOrderSet(FilterSet):
    meta = MetaFieldOrdering()
    answer_value = AnswerValueOrdering()
    attribute = AttributeOrderingFactory(
        models.Document, fields=["created_at", "modified_at", "form"]
    )

    class Meta:
        model = models.Document
        fields = ("meta", "answer_value", "attribute")


class VisibleAnswerFilter(Filter):
    field_class = BooleanField

    def filter(self, qs, value):
        if value in EMPTY_VALUES or not qs.exists():  # pragma: no cover
            # This prevents initializing the whole document structure with JEXL
            # computations if there is no value or no queryset to avoid useless
            # expensive computations
            return qs

        # assuming qs can only ever be in the context of a single document
        document = qs.first().document.family
        validator = validators.DocumentValidator()
        return qs.filter(question__slug__in=validator.visible_questions(document))


class VisibleQuestionFilter(Filter):
    field_class = GlobalIDFormField

    def filter(self, qs, value):
        if value in EMPTY_VALUES or not qs.exists():  # pragma: no cover
            return qs

        document_id = extract_global_id(value)

        # assuming qs can only ever be in the context of a single document
        document = models.Document.objects.get(pk=document_id)
        validator = validators.DocumentValidator()
        return qs.filter(slug__in=validator.visible_questions(document))


class QuestionFilterSet(MetaFilterSet):
    exclude_forms = GlobalIDMultipleChoiceFilter(field_name="forms", exclude=True)
    search = SearchFilter(fields=("slug", "label"))
    slugs = MultipleChoiceFilter(field_name="slug")

    visible_in_document = VisibleQuestionFilter()

    class Meta:
        model = models.Question
        fields = (
            "label",
            "is_required",
            "is_hidden",
            "is_archived",
            "sub_form",
            "row_form",
        )


class AnswerFilterSet(MetaFilterSet):
    search = SearchFilter(fields=("value", "file__name"))
    questions = GlobalIDMultipleChoiceFilter(field_name="question")

    visible_in_context = VisibleAnswerFilter()

    class Meta:
        model = models.Answer
        fields = ("question", "search")


class AnswerOrderSet(FilterSet):
    meta = MetaFieldOrdering()
    attribute = AttributeOrderingFactory(
        models.Answer,
        fields=[
            "created_at",
            "modified_at",
            "question",
            "value",
            "date",
        ],
    )

    class Meta:
        model = models.Answer
        fields = ("meta", "attribute")


class DynamicOptionFilterSet(FilterSet):
    question = GlobalIDFilter()
    document = GlobalIDFilter()

    class Meta:
        model = models.DynamicOption
        fields = ("question", "document")


class DynamicOptionOrderSet(FilterSet):
    meta = MetaFieldOrdering()
    attribute = AttributeOrderingFactory(
        models.DynamicOption,
        fields=["created_at", "modified_at", "slug", "label", "question"],
    )

    class Meta:
        model = models.DynamicOption
        fields = ("meta", "attribute")



import enum

import graphene
from django.core.exceptions import ImproperlyConfigured
from django.utils import translation
from graphene_django.registry import get_global_registry
from graphene_django.rest_framework import serializer_converter
from graphql_relay import to_global_id
from localized_fields.fields import LocalizedField
from rest_framework import relations, serializers
from rest_framework.serializers import ChoiceField

from .jexl import JexlValidator
from .relay import extract_global_id


class CalumaChoiceField(ChoiceField):
    """Custom choice field.

    Correctly handles Enum values that graphene parses before the
    value gets to the DRF serializer.
    """

    def to_internal_value(self, data):
        # TODO: This shouldn't be required IMHO - find out why
        # graphene parses the enum value before we get to it
        #
        # This is a workaround for the following bug:
        # https://github.com/graphql-python/graphene-django/issues/1280
        #
        # If/when this bug is fixed, this whole intermediate class may
        # become obsolete
        if isinstance(data, enum.Enum):
            data = data.value
        return super().to_internal_value(data)


class GlobalIDPrimaryKeyRelatedField(relations.PrimaryKeyRelatedField):
    """
    References a global id primary key.

    This class ensures that only primary keys may be looked up which are visible
    by corresponding DjangoObjectType.
    """

    def get_queryset(self):
        queryset = super().get_queryset()
        registry = get_global_registry()
        node_type = registry.get_type_for_model(queryset.model)
        return node_type.get_queryset(queryset, self.context.get("info"))

    def to_internal_value(self, data):
        data = extract_global_id(data)
        return super().to_internal_value(data)

    def to_representation(self, value):
        value = super().to_representation(value)
        return to_global_id(self.get_queryset().model.__name__, value)


class GlobalIDField(serializers.Field):
    def to_internal_value(self, data):
        return extract_global_id(data)


class JexlField(serializers.CharField):
    def __init__(self, jexl, **kwargs):
        super().__init__(**kwargs)
        self.validators.append(JexlValidator(jexl))


class ModelSerializer(serializers.ModelSerializer):
    serializer_related_field = GlobalIDPrimaryKeyRelatedField
    serializer_field_mapping = {
        **serializers.ModelSerializer.serializer_field_mapping,
        LocalizedField: serializers.CharField,
    }

    # will be set in caluma_core.AppConfig.ready hook, see apps.py
    # to avoid recursive import error
    validation_classes = None

    def validate(self, data):
        mutation = self.context["mutation"]
        info = self.context["info"]

        if self.validation_classes is None:
            raise ImproperlyConfigured(
                "check that app `caluma.caluma_core` is part of your `INSTALLED_APPS` "
                "or custom mutation has `validation_classes` properly assigned."
            )

        for validation_class in self.validation_classes:
            data = validation_class().validate(mutation, data, info)

        return data

    def create(self, validated_data):
        user = self.context["request"].user
        validated_data["created_by_user"] = user.username
        validated_data["created_by_group"] = user.group

        return super().create(validated_data)

    def update(self, instance, validated_data):
        user = self.context["request"].user
        validated_data["modified_by_user"] = user.username
        validated_data["modified_by_group"] = user.group

        return super().update(instance, validated_data)

    def build_standard_field(self, field_name, model_field):
        field_class, field_kwargs = super().build_standard_field(
            field_name, model_field
        )

        if isinstance(model_field, LocalizedField):
            lang = translation.get_language()
            allow_blank = model_field.blank or lang not in model_field.required
            field_kwargs["allow_blank"] = allow_blank

        return field_class, field_kwargs


@serializer_converter.get_graphene_type_from_serializer_field.register(
    serializers.ManyRelatedField
)
def convert_serializer_primary_key_related_field(field):
    return (graphene.List, graphene.ID)


@serializer_converter.get_graphene_type_from_serializer_field.register(
    relations.RelatedField
)
@serializer_converter.get_graphene_type_from_serializer_field.register(GlobalIDField)
def convert_serializer_relation_to_id(field):
    # TODO: could be removed once following issue is fixed
    # https://github.com/graphql-python/graphene-django/issues/389
    return graphene.ID



# Generated by Django 2.2.18 on 2021-03-04 14:11

from django.db import migrations, models
from django.db.models.expressions import F


def initial_modified_properties(apps, schema_editor):
    for model in [
        "Answer",
        "AnswerDocument",
        "Document",
        "DynamicOption",
        "File",
        "Form",
        "FormQuestion",
        "Option",
        "Question",
        "QuestionOption",
    ]:
        Model = apps.get_model(f"caluma_form.{model}")
        Model.objects.update(
            modified_by_user=F("created_by_user"),
            modified_by_group=F("created_by_group"),
        )


class Migration(migrations.Migration):

    dependencies = [
        ("caluma_form", "0039_merge_20210114_0956"),
    ]

    operations = [
        migrations.AddField(
            model_name="answer",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="answer",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="answerdocument",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="answerdocument",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="document",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="document",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="dynamicoption",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="dynamicoption",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="file",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="file",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="form",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="form",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="formquestion",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="formquestion",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalanswer",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalanswer",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalanswerdocument",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalanswerdocument",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicaldocument",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicaldocument",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicaldynamicoption",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicaldynamicoption",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalfile",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalfile",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalform",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalform",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalformquestion",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalformquestion",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicaloption",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicaloption",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalquestion",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalquestion",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalquestionoption",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="historicalquestionoption",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="option",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="option",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="question",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="question",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="questionoption",
            name="modified_by_group",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.AddField(
            model_name="questionoption",
            name="modified_by_user",
            field=models.CharField(
                blank=True, db_index=True, max_length=150, null=True
            ),
        ),
        migrations.RunPython(initial_modified_properties, migrations.RunPython.noop),
    ]



import uuid

from django.db import models
from simple_history.models import HistoricalRecords


def _history_user_getter(historical_instance):
    return historical_instance.history_user_id


def _history_user_setter(historical_instance, user):
    request = getattr(HistoricalRecords.thread, "request", None)
    user = None
    if request is not None:
        user = request.user.username
        if request.user.__class__.__name__ == "AnonymousUser":
            user = "AnonymousUser"
    historical_instance.history_user_id = user


class HistoricalModel(models.Model):
    history = HistoricalRecords(
        inherit=True,
        history_user_id_field=models.CharField(null=True, max_length=150),
        history_user_setter=_history_user_setter,
        history_user_getter=_history_user_getter,
    )

    class Meta:
        abstract = True


class BaseModel(models.Model):
    created_at = models.DateTimeField(auto_now_add=True, db_index=True)
    modified_at = models.DateTimeField(auto_now=True, db_index=True)
    created_by_user = models.CharField(
        max_length=150, blank=True, null=True, db_index=True
    )
    created_by_group = models.CharField(
        max_length=150, blank=True, null=True, db_index=True
    )
    modified_by_user = models.CharField(
        max_length=150, blank=True, null=True, db_index=True
    )
    modified_by_group = models.CharField(
        max_length=150, blank=True, null=True, db_index=True
    )

    def __str__(self):
        return str(self.pk)

    class Meta:
        abstract = True


class SlugModel(BaseModel, HistoricalModel):
    """
    Models which use a slug as primary key.

    Defined as Caluma default for configuration so it is possible
    to merge between developer and user configuration.
    """

    # Slug is limited to 127 chars to enable the NaturalKeyModel
    # (see below) to reference two slugs, separated with a dot.
    slug = models.SlugField(max_length=127, primary_key=True)

    def __repr__(self):
        return f"{self.__class__.__name__}(slug={self.slug})"

    class Meta:
        abstract = True


class UUIDModel(BaseModel, HistoricalModel):
    """
    Models which use uuid as primary key.

    Defined as Caluma default
    """

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    def __repr__(self):
        return f"{self.__class__.__name__}(id={self.id})"

    class Meta:
        abstract = True


class NaturalKeyModel(BaseModel, HistoricalModel):
    """Models which use a natural key as primary key."""

    id = models.CharField(max_length=255, unique=True, primary_key=True)

    def natural_key(self):  # pragma: no cover
        raise NotImplementedError()

    def save(self, *args, **kwargs):
        self.id = self.natural_key()
        return super().save(*args, **kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}(id={self.id})"

    class Meta:
        abstract = True


class ChoicesCharField(models.CharField):
    # This is only needed for backwards compatibility in the migrations
    pass



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-04-11 06:07
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0010_auto_20190404_0652")]

    operations = [
        migrations.AddField(
            model_name="answer",
            name="date",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="answer",
            name="value",
            field=models.JSONField(blank=True, null=True),
        ),
    ]



from typing import Any, Tuple, Union

from django import forms
from django.db.models.expressions import CombinedExpression, F, Value
from django.db.models.query import QuerySet
from django_filters.fields import ChoiceField
from django_filters.rest_framework import Filter
from graphene import Enum
from graphene_django.forms.converter import convert_form_field
from graphene_django.registry import get_global_registry
from rest_framework import exceptions

OrderingFieldType = Union[F, str]


class CalumaOrdering(Filter):
    """Base class for new-style ordering filters.

    This is not really a filter in the rest framework sense,
    but we use it's infrastructure to get conversion to GraphQL
    types.

    This is an abstract base class. To use it, you need to
    implement the `get_ordering_value()` method to return a tuple
    of `(queryset, ordering_field)`. The queryset may be modified,
    for example if you need to annotate it to get to the desired
    ordering field.
    """

    def get_ordering_value(
        self, qs: QuerySet, value: Any
    ) -> Tuple[QuerySet, OrderingFieldType]:  # pragma: no cover
        raise NotImplementedError("get_ordering_value() needs to be implemented")


class MetaFieldOrdering(CalumaOrdering):
    """Ordering filter for ordering by `meta` values."""

    def __init__(self, field_name="meta"):
        super().__init__()
        self.field_name = field_name

    field_class = forms.CharField

    def get_ordering_value(
        self, qs: QuerySet, value: Any
    ) -> Tuple[QuerySet, OrderingFieldType]:
        value = (hasattr(value, "value") and value.value) or value  # noqa: B009

        return qs, CombinedExpression(F(self.field_name), "->", Value(value))


class AttributeOrderingMixin(CalumaOrdering):
    def get_ordering_value(self, qs, value):
        value = (hasattr(value, "value") and value.value) or value  # noqa: B009
        if value not in self._fields:  # pragma: no cover
            # this is normally covered by graphene enforcing its schema,
            # but we still need to handle it
            raise exceptions.ValidationError(
                f"Field '{value}' not available for sorting on {qs.model.__name}"
            )
        return qs, F(value)


def AttributeOrderingFactory(model, fields):
    """Build ordering field for a given model.

    Used to define an ordering field that is presented as an enum
    """
    field_enum = type(
        f"Sortable{model.__name__}Attributes",
        (Enum,),
        {field.upper(): field for field in fields},
    )

    field_class = type(f"{model.__name__}AttributeField", (ChoiceField,), {})

    ordering_class = type(
        f"{model.__name__}AttributeOrdering",
        (AttributeOrderingMixin, CalumaOrdering),
        {"field_class": field_class, "_fields": fields},
    )

    @convert_form_field.register(field_class)
    def to_enum(field):
        registry = get_global_registry()
        converted = registry.get_converted_field(field)

        if converted:  # pragma: no cover
            return converted

        converted = field_enum()

        registry.register_converted_field(field, converted)
        return converted

    return ordering_class()



# Generated by Django 3.2.15 on 2022-11-25 16:22

from django.db import migrations


def rename_form_slug_fields(apps, schema_editor):
    AnalyticsField = apps.get_model("caluma_analytics.AnalyticsField")
    for field in AnalyticsField.objects.filter(data_source__endswith="form_id"):
        field.data_source = field.data_source.replace("form_id", "caluma_form.slug")
        field.save()


class Migration(migrations.Migration):

    dependencies = [
        ("caluma_analytics", "0005_analytics_field_ordering"),
    ]

    operations = [
        migrations.RunPython(rename_form_slug_fields, migrations.RunPython.noop)
    ]



# Generated by Django 3.2.12 on 2022-03-24 12:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("caluma_analytics", "0001_initial"),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name="analyticsfield",
            name="unique_data_source",
        ),
        migrations.AddField(
            model_name="analyticsfield",
            name="function",
            field=models.CharField(
                choices=[
                    ("value", "Plain value (Implies GROUP BY)"),
                    ("sum", "Sum"),
                    ("count", "Count"),
                    ("avg", "Average (Mean)"),
                    ("max", "Max"),
                    ("min", "Min"),
                ],
                default="value",
                max_length=20,
            ),
        ),
        migrations.AddField(
            model_name="analyticsfield",
            name="show_output",
            field=models.BooleanField(default=True),
        ),
        migrations.AddField(
            model_name="historicalanalyticsfield",
            name="function",
            field=models.CharField(
                choices=[
                    ("value", "Plain value (Implies GROUP BY)"),
                    ("sum", "Sum"),
                    ("count", "Count"),
                    ("avg", "Average (Mean)"),
                    ("max", "Max"),
                    ("min", "Min"),
                ],
                default="value",
                max_length=20,
            ),
        ),
        migrations.AddField(
            model_name="historicalanalyticsfield",
            name="show_output",
            field=models.BooleanField(default=True),
        ),
        migrations.AddConstraint(
            model_name="analyticsfield",
            constraint=models.UniqueConstraint(
                fields=("table", "data_source", "function"), name="unique_data_source"
            ),
        ),
    ]



# Generated by Django 2.2.16 on 2020-09-18 15:59

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="AccessLog",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                ("username", models.CharField(blank=True, max_length=100, null=True)),
                ("query", models.TextField()),
                (
                    "variables",
                    models.JSONField(blank=True, null=True),
                ),
                (
                    "operation_name",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
                ("status_code", models.PositiveIntegerField()),
                ("has_error", models.BooleanField(default=False)),
                (
                    "operation",
                    models.CharField(
                        choices=[("query", "query"), ("mutation", "mutation")],
                        max_length=20,
                    ),
                ),
                ("selection", models.CharField(blank=True, max_length=100, null=True)),
            ],
        )
    ]



# Generated by Django 2.2.15 on 2020-09-01 14:35

from django.db import migrations

import caluma.caluma_core.models


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0025_auto_20200901_1415")]

    operations = [
        migrations.AlterField(
            model_name="case",
            name="status",
            field=caluma.caluma_core.models.ChoicesCharField(
                choices=[
                    ("running", "Case is running and work items need to be completed."),
                    ("completed", "Case is done."),
                    ("canceled", "Case is canceled."),
                    ("suspended", "Case is suspended."),
                ],
                db_index=True,
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalcase",
            name="status",
            field=caluma.caluma_core.models.ChoicesCharField(
                choices=[
                    ("running", "Case is running and work items need to be completed."),
                    ("completed", "Case is done."),
                    ("canceled", "Case is canceled."),
                    ("suspended", "Case is suspended."),
                ],
                db_index=True,
                max_length=50,
            ),
        ),
    ]



from collections import defaultdict
from contextlib import contextmanager
from functools import partial

from pyjexl.analysis import ValidatingAnalyzer
from pyjexl.evaluator import Context

from ..caluma_core.jexl import (
    JEXL,
    ExtractTransformArgumentAnalyzer,
    ExtractTransformSubjectAnalyzer,
    ExtractTransformSubjectAndArgumentsAnalyzer,
)
from .models import Question
from .structure import Field


class QuestionMissing(Exception):
    pass


class QuestionValidatingAnalyzer(ValidatingAnalyzer):
    def visit_Transform(self, transform):
        if transform.name == "answer" and not isinstance(transform.subject.value, str):
            yield f"{transform.subject.value} is not a valid question slug."

        yield from super().visit_Transform(transform)


class QuestionJexl(JEXL):
    def __init__(self, validation_context=None, **kwargs):

        if validation_context:
            if "jexl_cache" not in validation_context:
                validation_context["jexl_cache"] = defaultdict(dict)
            self._cache = validation_context["jexl_cache"]
        else:
            self._cache = defaultdict(dict)

        super().__init__(**kwargs)

        self._structure = None
        self._form = None

        context_data = None

        if validation_context:
            # cleaned up variant
            self._form = validation_context.get("form")
            self._structure = validation_context.get("structure")
            context_data = {
                "form": self._form.slug if self._form else None,
                "info": self._structure,
            }

        self.context = Context(context_data)

        self.add_transform("answer", self.answer_transform)

    def answer_transform(self, question_slug, *args):
        field = self._structure.get_field(question_slug)

        # The first and only argument is the default value. If passed the field
        # is not required and we return that argument.
        if not field and len(args):
            return args[0]

        if self.is_hidden(field):
            return field.question.empty_value()

        # This overrides the logic in field.value() to consider visibility for
        # table cells
        elif field.question.type == Question.TYPE_TABLE and field.answer is not None:
            return [
                {
                    cell.question.slug: cell.value()
                    for cell in row.children()
                    if not self.is_hidden(cell)
                }
                for row in field.children()
            ]

        return field.value()

    def validate(self, expression, **kwargs):
        return super().validate(expression, QuestionValidatingAnalyzer)

    def extract_referenced_questions(self, expr):
        transforms = ["answer"]
        yield from self.analyze(
            expr, partial(ExtractTransformSubjectAnalyzer, transforms=transforms)
        )

    def extract_referenced_questions_with_arguments(self, expr):
        transforms = ["answer"]
        yield from self.analyze(
            expr,
            partial(ExtractTransformSubjectAndArgumentsAnalyzer, transforms=transforms),
        )

    def extract_referenced_mapby_questions(self, expr):
        transforms = ["mapby"]
        yield from self.analyze(
            expr, partial(ExtractTransformArgumentAnalyzer, transforms=transforms)
        )

    @contextmanager
    def use_field_context(self, field: Field):
        """Context manger to temporarily overwrite self._structure.

        This is used so we can evaluate each JEXL expression in the context
        of the corresponding question, not from where the question was
        referenced.
        This is relevant in table questions and form questions, so we always
        lookup the correct answer value (no "crosstalk" between rows, for example)
        """

        # field's parent is the fieldset - which is a valid structure object
        old_structure = self._structure
        self._structure = field.parent() or self._structure
        yield
        self._structure = old_structure

    def _get_referenced_fields(self, field: Field, expr: str):
        deps = list(self.extract_referenced_questions_with_arguments(expr))
        referenced_fields = [self._structure.get_field(slug) for slug, _ in deps]

        referenced_slugs = [ref.question.slug for ref in referenced_fields if ref]

        for slug, args in deps:
            required = len(args) == 0
            if slug not in referenced_slugs and required:
                raise QuestionMissing(
                    f"Question `{slug}` could not be found in form {field.form}"
                )

        return [field for field in referenced_fields if field]

    def is_hidden(self, field: Field):
        """Return True if the given field is hidden.

        This checks whether the dependency questions are hidden, then
        evaluates the field's is_hidden expression itself.
        """
        cache_key = (field.document.pk, field.question.pk)

        if cache_key in self._cache["hidden"]:
            return self._cache["hidden"][cache_key]

        # Check visibility of dependencies before actually evaluating the `is_hidden`
        # expression. If all dependencies are hidden,
        # there is no way to evaluate our own visibility, so we default to
        # hidden state as well.
        referenced_fields = self._get_referenced_fields(field, field.question.is_hidden)

        # all() returns True for the empty set, thus we need to
        # check that we have some deps at all first
        all_deps_hidden = bool(referenced_fields) and all(
            self.is_hidden(ref_field) for ref_field in referenced_fields
        )
        if all_deps_hidden:
            self._cache["hidden"][cache_key] = True
            return True

        # Also check if the question is hidden indirectly,
        # for example via parent formquestion.
        parent = field.parent()
        if parent and parent.question and self.is_hidden(parent):
            # no way this is shown somewhere
            self._cache["hidden"][cache_key] = True
            return True

        # if the question is visible-in-context and not hidden by invisible dependencies,
        # we can evaluate it's own is_hidden expression
        with self.use_field_context(field):
            self._cache["hidden"][cache_key] = self.evaluate(field.question.is_hidden)

        return self._cache["hidden"][cache_key]

    def is_required(self, field: Field):
        cache_key = (field.document.pk, field.question.pk)
        question = field.question

        if cache_key in self._cache["required"]:
            return self._cache["required"][cache_key]

        referenced_fields = self._get_referenced_fields(field, question.is_required)

        # all() returns True for the empty set, thus we need to
        # check that we have some deps at all first
        all_deps_hidden = bool(referenced_fields) and all(
            self.is_hidden(ref_field) for ref_field in referenced_fields
        )
        if all_deps_hidden:
            ret = False
        else:
            with self.use_field_context(field):
                ret = self.evaluate(question.is_required)
        self._cache["required"][cache_key] = ret
        return ret

    def evaluate(self, expr, raise_on_error=True):
        try:
            return super().evaluate(expr)
        except TypeError:
            if raise_on_error:
                raise
            return None



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-04-18 07:33
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0012_auto_20190416_1835")]

    operations = [
        migrations.AddField(
            model_name="question",
            name="data_source",
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name="question",
            name="type",
            field=models.CharField(
                choices=[
                    ("multiple_choice", "multiple_choice"),
                    ("integer", "integer"),
                    ("float", "float"),
                    ("date", "date"),
                    ("choice", "choice"),
                    ("textarea", "textarea"),
                    ("text", "text"),
                    ("table", "table"),
                    ("form", "form"),
                    ("file", "file"),
                    ("dynamic_choice", "dynamic_choice"),
                    ("dynamic_multiple_choice", "dynamic_multiple_choice"),
                ],
                max_length=23,
            ),
        ),
    ]



from graphene_django import forms


class GlobalIDFormField(forms.GlobalIDFormField):
    """
    Global id form field which allows plain primary keys.

    Disable cleaning as input may be an actual relay global id
    or plain primary key.
    """

    def clean(self, value):  # pragma: no cover
        return value


class GlobalIDMultipleChoiceField(forms.GlobalIDMultipleChoiceField):
    """
    Global id multiple choice field which allows plain primary keys.

    Disable cleaning as input may be an actual relay global id
    or plain primary key.
    """

    def valid_value(self, value):  # pragma: no cover
        return True



from __future__ import annotations  # self-referencing annotations

from collections import defaultdict
from functools import cached_property

from django.db import connection
from psycopg2.extras import DictCursor

from . import simple_table, sql

FUNCTION_PARAMETER_CAST = {
    "sum": "::float",
    "avg": "::float",
}


class PivotTable:
    class _AnonymousInfo:
        """
        Pseudo GraphQL info object.

        Used for commandline access, so the commandline version
        can also invoke the visibilities.
        """

        context = {}

    def __init__(self, table, info=_AnonymousInfo):
        self.info = info

        self.table = table
        self.last_query = None
        self.last_query_params = None
        self.base_table = simple_table.SimpleTable(self.table, info=info)
        self._summary = defaultdict(int)

    @cached_property
    def _fields(self):
        fields = self.table.fields.all()

        return list(fields)

    def get_sql_and_params(self):
        """Return a list of records as specified in the given table config."""

        group_by = []

        # We replace the selects from the analytics query with
        # the aggregate function calls as required, but leave the
        # rest intact. This way, we won't need to build a subquery
        # around the whole thing and keep it relatively (!) simple
        aggregate_query = self.base_table.get_query_object()

        selects_by_alias = {
            alias: fragment for fragment, alias in aggregate_query.select
        }

        new_selects = []

        for field in self._fields:
            key = f"analytics_result_{field.alias}"

            # old field might be directly selected in the base query,
            # in which case we need to alias it's expression. If it's
            # not directly in the base, it won't be in the top-level selects
            # list, but will be aliased to the known name in a subquery.
            old_field = selects_by_alias.get(key, key)

            new_alias = f"analytics_result_{field.alias}"

            if field.function == field.FUNCTION_VALUE:
                group_by.append(old_field)
                new_selects.append((old_field, new_alias))
            elif field.function != field.FUNCTION_VALUE:
                cast = FUNCTION_PARAMETER_CAST.get(field.function, "")
                new_selects.append((f"{field.function}({old_field}{cast})", new_alias))

        # TODO: instead of replacing selects, use sub-query
        aggregate_query.select = new_selects
        aggregate_query.group_by = group_by
        aggregate_query.select_direct_only = True

        sql_query, params, _ = sql.QueryRender(aggregate_query).as_sql(alias=None)

        return sql_query, params

    @cached_property
    def field_ordering(self):
        return list(self.table.fields.all().values_list("alias", flat=True))

    def get_records(self):
        self._summary = defaultdict(int)
        sql_query, params = self.get_sql_and_params()

        with connection.connection.cursor(cursor_factory=DictCursor) as cursor:
            cursor.execute(sql_query, params)
            data = cursor.fetchall()

            result = []
            for row in data:
                record = {}
                for field in self._fields:
                    field2 = self.base_table._fields[field.alias]
                    if field.show_output:
                        value = row[f"analytics_result_{field.alias}"]
                        self._update_summary(field, value)
                        record[field.alias] = field2.parse_value(value)
                result.append(record)

            return result

    def _update_summary(self, field, value):
        # value must to eval to True in order to not break on `None` and also ignore `0`
        if value and field.function in [field.FUNCTION_SUM, field.FUNCTION_COUNT]:
            self._summary[field.alias] += value

    def get_summary(self):
        if not self._summary:  # pragma: no cover
            self.get_records()
        return self._summary

    def _sql_alias(self, user_alias):  # pragma: no cover
        return f"analytics_pivot_{user_alias}"



# -*- coding: utf-8 -*-
# Generated by Django 1.11.17 on 2018-12-12 12:43
from __future__ import unicode_literals

import uuid

import django.db.models.deletion
import localized_fields.fields.field
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Answer",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("value", models.JSONField()),
                ("meta", models.JSONField(default={})),
            ],
        ),
        migrations.CreateModel(
            name="Document",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("meta", models.JSONField(default={})),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="Form",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                ("slug", models.SlugField(primary_key=True, serialize=False)),
                ("name", localized_fields.fields.field.LocalizedField(required=[])),
                (
                    "description",
                    localized_fields.fields.field.LocalizedField(
                        blank=True, null=True, required=[]
                    ),
                ),
                ("meta", models.JSONField(default={})),
                ("is_published", models.BooleanField(default=False)),
                ("is_archived", models.BooleanField(default=False)),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="FormQuestion",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "sort",
                    models.PositiveIntegerField(
                        db_index=True, default=0, editable=False
                    ),
                ),
                (
                    "form",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="caluma_form.Form",
                    ),
                ),
            ],
            options={"ordering": ("-sort", "id")},
        ),
        migrations.CreateModel(
            name="Option",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                ("slug", models.SlugField(primary_key=True, serialize=False)),
                ("label", localized_fields.fields.field.LocalizedField(required=[])),
                ("meta", models.JSONField(default={})),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="Question",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                ("slug", models.SlugField(primary_key=True, serialize=False)),
                ("label", localized_fields.fields.field.LocalizedField(required=[])),
                (
                    "type",
                    models.CharField(
                        choices=[
                            ("checkbox", "checkbox"),
                            ("integer", "integer"),
                            ("float", "float"),
                            ("radio", "radio"),
                            ("textarea", "textarea"),
                            ("text", "text"),
                        ],
                        max_length=10,
                    ),
                ),
                ("is_required", models.TextField(default="false")),
                ("is_hidden", models.TextField(default="false")),
                ("is_archived", models.BooleanField(default=False)),
                (
                    "configuration",
                    models.JSONField(default={}),
                ),
                ("meta", models.JSONField(default={})),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="QuestionOption",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "sort",
                    models.PositiveIntegerField(
                        db_index=True, default=0, editable=False
                    ),
                ),
                (
                    "option",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="caluma_form.Option",
                    ),
                ),
                (
                    "question",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="caluma_form.Question",
                    ),
                ),
            ],
            options={"ordering": ("-sort", "id")},
        ),
        migrations.AddField(
            model_name="question",
            name="options",
            field=models.ManyToManyField(
                related_name="questions",
                through="caluma_form.QuestionOption",
                to="caluma_form.Option",
            ),
        ),
        migrations.AddField(
            model_name="formquestion",
            name="question",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="caluma_form.Question"
            ),
        ),
        migrations.AddField(
            model_name="form",
            name="questions",
            field=models.ManyToManyField(
                related_name="forms",
                through="caluma_form.FormQuestion",
                to="caluma_form.Question",
            ),
        ),
        migrations.AddField(
            model_name="document",
            name="form",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="documents",
                to="caluma_form.Form",
            ),
        ),
        migrations.AddField(
            model_name="answer",
            name="document",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="answers",
                to="caluma_form.Document",
            ),
        ),
        migrations.AddField(
            model_name="answer",
            name="question",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="answers",
                to="caluma_form.Question",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="questionoption", unique_together=set([("option", "question")])
        ),
        migrations.AlterUniqueTogether(
            name="formquestion", unique_together=set([("form", "question")])
        ),
        migrations.AlterUniqueTogether(
            name="answer", unique_together=set([("document", "question")])
        ),
    ]



import logging

from caluma.caluma_form.models import DynamicOption
from caluma.utils import is_iterable_and_no_string

logger = logging.getLogger(__name__)


class BaseDataSource:
    """Basic data source class to be extended by any data source implementation.

    The `get_data`-method should return an iterable. This iterable can contain strings,
    ints, floats and also iterables. Those contained iterables can consist of maximally
    two items. The first will be used for the option name, the second one for it's
    value. If only one value is provided, this value will also be used as choice name.

    The `validate_answer_value`-method checks if each value in
    `self.get_data(user, question, context)` equals the value of the parameter
    `value`. If this is correct the method returns the label as a String and
    otherwise the method returns `False`.

    Examples:
        [['my-option', {"en": "english description", "de": "deutsche Beschreibung"}, ...]
        [['my-option', "my description"], ...]
        ['my-option', ...]
        [['my-option'], ...]

    Properties:
        info: Informational string about this data source
        default: default value to return if execution of `get_data()` fails.
                 If this is `None`, the Exception won't be handled. Defaults to None.

    A custom data source class could look like this:
    ```
    >>> from caluma.caluma_data_source.data_sources import BaseDataSource
    ... from caluma.caluma_data_source.utils import data_source_cache
    ... import requests
    ...
    ...
    ... class CustomDataSource(BaseDataSource):
    ...     info = 'User choices from "someapi"'
    ...
    ...     @data_source_cache(timeout=3600)
    ...     def get_data(self, user, question, context):
    ...         response = requests.get(f"https://someapi/?user={user.username}")
    ...         return [result["value"] for result in response.json()["results"]]
    ```

    """

    info = None
    default = None

    def __init__(self):
        pass

    def get_data(self, user, question, context):  # pragma: no cover
        raise NotImplementedError()

    def validate_answer_value(self, value, document, question, user, context):
        for data in self.get_data(user, question, context):
            label = data
            if is_iterable_and_no_string(data):
                label = data[-1]
                data = data[0]
            if str(data) == value:
                if not isinstance(label, dict):
                    label = str(label)
                return label
        dynamic_option = DynamicOption.objects.filter(
            document=document, question=question, slug=value
        ).first()
        if dynamic_option:
            return dynamic_option.label
        return False

    def try_get_data_with_fallback(self, user, question, context):
        try:
            new_data = self.get_data(user, question, context)
        except Exception as e:
            logger.exception(
                f"Executing {type(self).__name__}.get_data() failed:"
                f"{e}\n Using default data."
            )
            if self.default is None:
                raise e
            return self.default
        return new_data



from __future__ import annotations  # self-referencing annotations

import re
from copy import deepcopy
from datetime import datetime
from decimal import Decimal
from functools import cached_property, partial
from typing import List, Optional

from django.conf import settings
from django.db import connection
from django.utils import timezone, translation
from psycopg2.extras import DictCursor

from caluma.caluma_form import models as form_models
from caluma.caluma_workflow import models as workflow_models

from . import models, sql


class BaseField:
    """Base class for all (available) fields."""

    def __init__(
        self,
        *,
        parent: Optional[BaseField],
        identifier: str,
        visibility_source,
        label=None,
        **_,
    ):
        """Initialize new field instance.

        Parameters are as follows:
         - `parent`: an optional reference to the superordinate field
         - `identifier`: the DB field that's being extracted
         - `visibility_source`: Helper to fetch visibility-aware querysets
         - `label`: Optional label
        """
        self.parent = parent
        self.identifier = identifier
        self.visibility_source = visibility_source
        self.label = label if label else identifier
        self.filters = None

        # required - set to True if the result of this field is
        # needed in the analysis.
        # Optimisation - SQL generation may be skipped in some
        # cases if field not required
        self.required = False
        self.show_output = True

    def path_args(self):
        """Extract arguments from the field's path."""
        match = re.match(r"^\w+?\[(.*)\]$", self.identifier)
        if match:
            return match.group(1).split(",")
        return []  # pragma: no cover

    @property
    def available_children(self):  # pragma: no cover
        """Return a dict of all available children.

        Used in the construction/design-time stage,
        as well as when building an SQL query from a given
        set of paths.
        """
        return {}

    def full_label(self):
        """Return a full hierarchical path to the given field.

        Includes the full label of the parent field(s) as well
        """
        if self.parent:
            return f"{self.parent.full_label()} > {self.label}"
        return self.label

    def source_path(self) -> List[str]:
        """Return the full source path of this field as a list."""
        return (
            self.parent.source_path() + [self.identifier]
            if self.parent
            else [self.identifier]
        )

    def parse_value(self, value):
        """Parse value from DB into python value.

        Ensure date / time types are using the correct timezone,
        and numeric types are int or float.
        """
        if isinstance(value, Decimal):
            # Postgres sometimes returns Decimal types that
            # need to be converted for "proper" representation
            # in the output. Depends on PostgreSQL version
            # whether this is triggered or not
            return float(value)  # pragma: no cover

        if isinstance(value, datetime):
            # psycopg2 returns "generic" tz info dates that sometimes
            # are weird when comparing to the ones Django expects
            # (Note they're still correct, just not labeled in a
            # useful way)
            current_tz = timezone.get_current_timezone()
            return current_tz.normalize(value)

        return value

    def query_field(self):
        """Return the corresponding SQL rendering field."""
        raise NotImplementedError(
            f"Method 'query_field' is missing in {type(self)}"
        )  # pragma: no cover

    def is_value(self):
        """Return True if this field represents a usable value."""
        raise NotImplementedError(
            f"Method 'is_value' is missing in {type(self)}"
        )  # pragma: no cover

    def is_leaf(self):
        """Return True if there are no more fields below this one."""
        raise NotImplementedError(
            f"Method 'is_leaf' is missing in {type(self)}"
        )  # pragma: no cover

    def supported_functions(self):
        """Return list of supported aggregate functions.

        In all cases, at least VALUE should be returned. Depending on
        data type of the field, additional functions can be supported.
        """
        raise NotImplementedError(
            f"Method 'supported_functions' is missing in {type(self)}"
        )  # pragma: no cover

    def _all_supported_functions(self):
        """Return a list of all functions that Analytics supports.

        call this in supported_functions() implementations if there
        is no restriction on supported functions in the field.
        """
        return [func.upper() for func, _ in models.AnalyticsField.FUNCTION_CHOICES]


class MetaField(BaseField):
    """MetaField implements access to JSON fields and their contents.

    For this to work correctly, the setting META_FIELDS must be
    configured to list the fields that can be extracted.
    If a key is set in a `meta` value but not listed in the
    settings, it will not be usable in Analytics.
    """

    def __init__(self, parent, identifier, label, *, visibility_source, meta_name=None):
        super().__init__(
            parent=parent,
            identifier=identifier,
            label=label,
            visibility_source=visibility_source,
        )
        self.meta_name = meta_name

    def is_leaf(self):
        # if we have a meta field name, we are a leaf:
        # We cannot descend further down, there are no children
        return bool(self.meta_name)

    def supported_functions(self):
        # meta is kind of special - we don't know what type of data people
        # put into those fields, so we support all functions despite some
        # of them might not work
        return self._all_supported_functions()

    def is_value(self):
        # "plain" meta cannot be used as a value
        return self.is_leaf()

    @cached_property
    def available_children(self):
        if self.meta_name:
            return {}
        return {
            name: MetaField(
                parent=self,
                identifier=self.identifier,
                label=f"Key '{name}'",
                meta_name=name,
                visibility_source=self.visibility_source,
            )
            for name in settings.META_FIELDS
        }

    def source_path(self):
        source_path = super().source_path()
        if self.meta_name:
            # replace last element with our "meta" name
            # so path is correct, and identifier still can point
            # to the actual DB field.
            source_path[-1] = self.meta_name
        return source_path

    def query_field(self):
        if self.meta_name:
            return sql.JSONExtractorField(
                self.identifier,
                self.identifier,
                parent=self.parent.query_field() if self.parent else None,
                json_key=self.meta_name,
            )
        return sql.NOOPField(
            self.identifier,
            self.identifier,
            parent=self.parent.query_field() if self.parent else None,
        )


class AttributeField(BaseField):
    """Extractor for regular table attributes.

    When `is_date` is set, it will also contain a number of sub-fields
    that represent extracts of some date parts (such as year, quarter,
    month, and weekday).
    """

    def __init__(self, parent, identifier, *, visibility_source, is_date=False):
        super().__init__(
            parent=parent, identifier=identifier, visibility_source=visibility_source
        )
        self.is_date = is_date

    def is_leaf(self):
        return not self.is_date

    def is_value(self):
        # meta fields cannot be a value - they *contain* values
        return self.identifier != "meta"

    def supported_functions(self):
        if not self.is_value:  # pragma: no cover
            # Non-value fields cannot have functions applied to them.
            # This mainly applies to "meta" fields in this context
            return []
        elif self.identifier == "id" or self.identifier.endswith("_id"):
            # Identifiers cannot have aggregate functions applied in a
            # meaningful manner, except counts. Potentially, we *could"
            # imagine some string functions being useful for slug identifiers,
            # but we'll keep it simple for now
            return [
                models.AnalyticsField.FUNCTION_COUNT.upper(),
                models.AnalyticsField.FUNCTION_VALUE.upper(),
            ]
        return [
            models.AnalyticsField.FUNCTION_VALUE.upper(),
            models.AnalyticsField.FUNCTION_MAX.upper(),
            models.AnalyticsField.FUNCTION_MIN.upper(),
            models.AnalyticsField.FUNCTION_COUNT.upper(),
        ]

    def query_field(self):
        if self.is_date and not self.required:
            return sql.NOOPField(
                self.identifier,
                self.identifier,
                parent=self.parent.query_field() if self.parent else None,
            )
        return sql.AttrField(
            self.identifier,
            self.identifier,
            parent=self.parent.query_field() if self.parent else None,
        )

    @cached_property
    def available_children(self):
        if self.is_date:
            # for FormAnswerField, the identifier is the question slug,
            # but the field is 'date'
            date_field = getattr(self, "attr_name", self.identifier)

            extractor_field = partial(
                DateExtractorField,
                parent=self,
                date_field=date_field,
                visibility_source=self.visibility_source,
            )
            return {
                "year": extractor_field(identifier="year"),
                "month": extractor_field(identifier="month"),
                "weekday": extractor_field(identifier="weekday"),
                "quarter": extractor_field(identifier="quarter"),
            }
        return {}


class WorkItemField(BaseField):
    """Field that provides access to a work item's data."""

    ORDER_TYPES = {
        "first": ["created_at"],
        "last": ["created_at DESC"],
        "firstclosed": ["closed_at"],
        "lastclosed": ["closed_at DESC"],
    }
    ORDER_LABELS = {
        "first": "first created",
        "last": "last created",
        "firstclosed": "first closed",
        "lastclosed": "first closed",
    }

    def __init__(
        self,
        parent,
        identifier,
        visibility_source,
        task_slug=None,
        ordering_selector=None,
    ):
        super().__init__(
            parent=parent, identifier=identifier, visibility_source=visibility_source
        )

        self.task_slug = task_slug if task_slug else self.path_args()[0]
        self.ordering_selector = (
            ordering_selector if ordering_selector else self.path_args()[1]
        )
        assert self.ordering_selector in self.ORDER_TYPES

        # TODO translate?
        self.label = f"Workitem: {self.ORDER_LABELS[self.ordering_selector]} with task {self.task_slug}"

    def _order_by(self):
        return self.ORDER_TYPES[self.ordering_selector]

    def is_leaf(self):
        return False

    def is_value(self):
        return False

    def supported_functions(self):
        return []

    def query_field(self):
        return sql.JoinField(
            self.identifier,
            self.identifier,
            table=self.visibility_source.work_items(),
            outer_ref=("id", "case_id"),
            filters=[f"task_id = '{self.task_slug}'"],
            order_by=self._order_by(),
            parent=self.parent.query_field() if self.parent else None,
        )

    @cached_property
    def available_children(self):
        direct_fields = {
            "meta": MetaField(
                parent=self,
                identifier="meta",
                label="Meta",
                visibility_source=self.visibility_source,
            ),
            "closed_at": AttributeField(
                parent=self,
                identifier="closed_at",
                is_date=True,
                visibility_source=self.visibility_source,
            ),
            "deadline": AttributeField(
                parent=self,
                identifier="closed_at",
                is_date=True,
                visibility_source=self.visibility_source,
            ),
            "created_at": AttributeField(
                parent=self,
                identifier="created_at",
                is_date=True,
                visibility_source=self.visibility_source,
            ),
            "status": AttributeField(
                parent=self,
                identifier="status",
                visibility_source=self.visibility_source,
            ),
            "name": AttributeField(
                parent=self, identifier="name", visibility_source=self.visibility_source
            ),
            "document_form": FormInfoField(
                parent=None,
                identifier="form",
                visibility_source=self.visibility_source,
            ),
        }

        wi_forms = form_models.Form.objects.filter(
            documents__work_item__isnull=False
        ).distinct()

        form_fields = {
            f"document[{form.slug}]": FormDocumentField(
                identifier=f"document[{form.slug}]",
                parent=self,
                visibility_source=self.visibility_source,
            )
            for form in wi_forms
        }

        return {**direct_fields, **form_fields}


class FormDocumentField(BaseField):
    """Field that provides access to a form's data."""

    def __init__(
        self, parent, identifier, visibility_source, subform_level=0, form_slug=None
    ):
        super().__init__(
            parent=parent, identifier=identifier, visibility_source=visibility_source
        )

        self.form_slug = form_slug if form_slug else self.path_args()[0]
        if self.form_slug == "*":
            self.form_slug = None

        # subform level is used so we can have the "path" as it is presented
        # to the user (our `location`) separate from where the corresponding
        # `Answer` is found (remember, answers of questions in subforms are
        # NOT put into sub-documents according to hierarchy, but stored in a
        # single document!)
        self.subform_level = subform_level

    def query_field(self):
        # Only for top-level form do we need a Join field.
        # The "sub-forms"' answers are all on the same document,
        # thus for those we return a NOOP field.
        if self.subform_level == 0:
            return sql.JoinField(
                self.identifier,
                self.identifier,
                table=self.visibility_source.documents(),
                outer_ref=("document_id", "id"),
                filters=[f"form_id = '{self.form_slug}'"] if self.form_slug else [],
                parent=self.parent.query_field() if self.parent else None,
            )
        # else - we just return the field from our "virtual" parent
        return self.parent.query_field()

    def is_leaf(self):
        return False

    def is_value(self):
        return False

    def supported_functions(self):
        return []

    @cached_property
    def available_children(self):
        if self.form_slug:
            form = form_models.Form.objects.get(pk=self.form_slug)
            questions = form.questions.all().exclude(
                type=form_models.Question.TYPE_TABLE
            )
        else:
            questions = self.visibility_source.questions(as_queryset=True).exclude(
                type=form_models.Question.TYPE_TABLE
            )

        children = {}
        if self.subform_level == 0:
            children["caluma_form"] = FormInfoField(
                parent=self,
                identifier="caluma_form",
                visibility_source=self.visibility_source,
            )

        for question in questions:
            qf = self._question_field(question)
            if qf:
                children[question.pk] = qf
        return children

    def _question_field(self, question):
        # Note: we explicitly don't support table questions at the moment.
        # They would require another sub-selector to get the first, or last
        # item in the table, which in most cases in real-world use isn't very
        # useful.
        if question.type in (
            question.TYPE_INTEGER,
            question.TYPE_FLOAT,
            question.TYPE_TEXT,
            question.TYPE_TEXTAREA,
            question.TYPE_CHOICE,
            question.TYPE_DYNAMIC_CHOICE,
        ):
            return FormAnswerField(
                parent=self,
                identifier=question.slug,
                attr_name="value",
                subform_level=self.subform_level,
                visibility_source=self.visibility_source,
            )
        elif question.type == question.TYPE_DATE:
            return FormAnswerField(
                parent=self,
                identifier=question.slug,
                attr_name="date",
                is_date=True,
                subform_level=self.subform_level,
                visibility_source=self.visibility_source,
            )

        elif question.type == question.TYPE_FORM:
            return FormDocumentField(
                parent=self,
                identifier=question.slug,
                form_slug=question.sub_form_id,
                subform_level=self.subform_level + 1,
                visibility_source=self.visibility_source,
            )


class DirectDocumentField(FormDocumentField):
    """Field that directly represents a document on top level.

    This is used only for the "Documents" starting object, as here,
    the main table is already the "document".
    """

    def query_field(self):
        return sql.NOOPField(
            self.identifier,
            self.identifier,
            parent=self.parent.query_field() if self.parent else None,
        )


class FormAnswerField(AttributeField):
    """Represents access to an answer within a form."""

    def __init__(self, parent, identifier, attr_name, subform_level, **kwargs):
        super().__init__(parent=parent, identifier=identifier, **kwargs)
        self.subform_level = subform_level
        self.attr_name = attr_name
        self._question = form_models.Question.objects.get(pk=self.identifier)

    def supported_functions(self):
        base_functions = [
            models.AnalyticsField.FUNCTION_VALUE.upper(),
            models.AnalyticsField.FUNCTION_COUNT.upper(),
        ]
        text_functions = [
            models.AnalyticsField.FUNCTION_MAX.upper(),
            models.AnalyticsField.FUNCTION_MIN.upper(),
        ]
        numeric_functions = [
            models.AnalyticsField.FUNCTION_MAX.upper(),
            models.AnalyticsField.FUNCTION_MIN.upper(),
            models.AnalyticsField.FUNCTION_AVERAGE.upper(),
            models.AnalyticsField.FUNCTION_SUM.upper(),
        ]
        function_support = {
            form_models.Question.TYPE_CHOICE: text_functions,
            form_models.Question.TYPE_TEXT: text_functions,
            form_models.Question.TYPE_TEXTAREA: text_functions,
            form_models.Question.TYPE_DATE: text_functions,
            form_models.Question.TYPE_INTEGER: numeric_functions,
            form_models.Question.TYPE_FLOAT: numeric_functions,
        }

        return base_functions + function_support[self._question.type]

    def query_field(self):
        # Only for top-level form do we need a Join field.
        # The "sub-forms"' answers are all on the same document,
        # thus for those we return a NOOP field.
        answer_join_field = sql.JoinField(
            identifier=self.identifier,
            extract=self.identifier,
            table=self.visibility_source.answers(),
            # TODO: turn this into SQL parameter
            filters=[f""""question_id" = '{self.identifier}' """],
            outer_ref=("id", "document_id"),
            parent=self.parent.query_field() if self.parent else None,
        )
        value_field = sql.AttrField(
            identifier=self.attr_name,
            extract=self.attr_name,
            parent=answer_join_field,
            filter_values=self.filters,
            answer_value_mode=(self.attr_name == "value"),
        )
        return value_field

    def is_leaf(self):
        if self._question.type == form_models.Question.TYPE_CHOICE:
            return False
        return super().is_leaf()

    def is_value(self):
        return (
            self._question.type == form_models.Question.TYPE_CHOICE
            or super().is_value()
        )

    @cached_property
    def available_children(self):
        if self._question.type == form_models.Question.TYPE_CHOICE:
            return {
                "label": ChoiceLabelField(
                    parent=self,
                    identifier="label",
                    visibility_source=self.visibility_source,
                )
            }
        return super().available_children


class FormInfoField(BaseField):
    """Represent information about a form (Name, slug)."""

    def __init__(self, parent, identifier, *, visibility_source, **kwargs):

        super().__init__(
            parent=parent,
            identifier=identifier,
            visibility_source=visibility_source,
            **kwargs,
        )

    def supported_functions(self):  # pragma: no cover
        return []

    def query_field(self):
        return sql.JoinField(
            identifier=self.identifier,
            extract=self.identifier,
            table=self.visibility_source.forms(),
            filters=[],
            outer_ref=("form_id", "slug"),
            parent=self.parent.query_field() if self.parent else None,
        )

    def is_leaf(self):
        return False

    def is_value(self):
        return False

    @cached_property
    def available_children(self):
        return {
            "name": LocalizedAttributeField(
                parent=self,
                identifier="name",
                visibility_source=self.visibility_source,
            ),
            "slug": AttributeField(
                parent=self,
                identifier="slug",
                is_date=False,
                visibility_source=self.visibility_source,
            ),
        }


class ChoiceLabelField(AttributeField):
    def __init__(
        self, parent, identifier, *, visibility_source, language=None, main_field=None
    ):
        super().__init__(
            parent=parent, identifier=identifier, visibility_source=visibility_source
        )
        self.language = language
        self._main_field = main_field

    def is_leaf(self):
        return bool(self.language)

    def is_value(self):  # pragma: no cover
        # without language, we return the default language as
        # per request, so we can still "be" a value
        return True

    def source_path(self) -> List[str]:
        """Return the full source path of this field as a list."""
        fragment = (
            [self._main_field.identifier, self.identifier]
            if self._main_field
            else [self.identifier]
        )
        return self.parent.source_path() + fragment if self.parent else fragment

    def query_field(self):
        # The label is in the choices table, so we need to join
        # that one.
        label_join_field = sql.JoinField(
            identifier=self.identifier,
            extract=self.identifier,
            table=self.visibility_source.options(),
            filters=[],
            outer_ref=("value #>>'{}'", "slug"),  # noqa:P103
            parent=self.parent.query_field() if self.parent else None,
        )

        language = self.language or translation.get_language()

        value_field = sql.HStoreExtractorField(
            "label",
            "label",
            parent=label_join_field,
            hstore_key=language,
        )

        return value_field

    @cached_property
    def available_children(self):
        if self.language:
            return {}
        return {
            lang: ChoiceLabelField(
                self.parent,
                lang,
                main_field=self,
                visibility_source=self.visibility_source,
                language=lang,
            )
            for lang, _ in settings.LANGUAGES
        }


class LocalizedAttributeField(AttributeField):
    def __init__(
        self, parent, identifier, *, visibility_source, language=None, main_field=None
    ):
        super().__init__(
            parent=parent, identifier=identifier, visibility_source=visibility_source
        )
        self.language = language
        self._main_field = main_field

    def is_leaf(self):
        return bool(self.language)

    def is_value(self):  # pragma: no cover
        # without language, we return the default language as
        # per request, so we can still "be" a value
        return True

    def source_path(self) -> List[str]:
        """Return the full source path of this field as a list."""
        fragment = (
            [self._main_field.identifier, self.identifier]
            if self._main_field
            else [self.identifier]
        )
        return self.parent.source_path() + fragment if self.parent else fragment

    def query_field(self):
        # The label is in the choices table, so we need to join
        # that one.

        language = self.language or translation.get_language()

        value_field = sql.HStoreExtractorField(
            "name",
            "name",
            parent=self.parent.query_field() if self.parent else None,
            hstore_key=language,
        )

        return value_field

    @cached_property
    def available_children(self):
        if self.language:
            return {}
        return {
            lang: LocalizedAttributeField(
                self.parent,
                lang,
                main_field=self,
                visibility_source=self.visibility_source,
                language=lang,
            )
            for lang, _ in settings.LANGUAGES
        }


class DateExtractorField(AttributeField):
    """Specialisation of attribute field: Extract a date part."""

    def __init__(self, parent, identifier, visibility_source, date_field, **kwargs):
        super().__init__(
            parent,
            identifier,
            visibility_source=visibility_source,
            is_date=False,
            **kwargs,
        )
        self.date_field = date_field

    def supported_functions(self):
        return self._all_supported_functions()

    def query_field(self):
        return sql.DateExprField(
            self.identifier,
            self.date_field,
            parent=self.parent.query_field() if self.parent else None,
            extract_part=self.identifier,
            filter_values=self.filters,
        )

    def parse_value(self, value):
        """Convert extracted date part to integer.

        This is necessary because not all supported postgres versions return the
        same datatype which leads to inconsistency. E.g versions < 12 returns a
        float while later versions return a decimal.
        """
        return int(value)


class BaseStartingObject:
    """Base class for starting objects.

    Provides interface for listing the fields within the given starting
    object, and also supports Enum generation on the GraphQL side.
    """

    def __init__(self, info, disable_visibilities):
        """Initialize starting object.

        The `info` parameter is the Graphql info/context object.
        """
        self.info = info

        # deferred import to avoid circular deps
        from . import visibility

        self.visibility_source = visibility.CalumaVisibilitySource(
            info, is_disabled=disable_visibilities
        )

    @classmethod
    def get_object(cls, identifier, info, disable_visibilities=False):
        """Return starting object identified by given name.

        Call `BaseStartingObject.get_object('cases', info)`
        to get a CaseStartingObject, for example.
        """

        subclasses_by_name = {sub.identifier: sub for sub in cls.__subclasses__()}
        return subclasses_by_name[identifier](
            info, disable_visibilities=disable_visibilities
        )

    @classmethod
    def as_choices(cls):
        """Return a list of (slug, label) tuples of all available starting objects.

        This can then be used in a CharField's choices.
        """

        objects = sorted(
            cls.__subclasses__(),
            key=lambda obj: obj.label,
        )
        return [(obj.identifier, obj.label) for obj in objects]

    def get_fields(self, prefix=None, depth=1):
        """Return fields for the GraphQL interface.

        When not parametrized, return the root fields of the starting object.

        When a given prefix and depth are given, return the fields from the
        specified sub-path, and go down the requested depth.
        """
        if prefix:
            # support string prefixes as well
            prefix = prefix if isinstance(prefix, list) else prefix.split(".")

        def _children_at_depth(field, _depth):
            if _depth <= 0:
                return
            for path, child in field.available_children.items():
                yield ([path], child)
                for subpath, grandchild in _children_at_depth(child, _depth - 1):
                    yield ([path] + subpath, grandchild)

        if prefix:
            # get only fields with given prefix, and start looking
            # down `depth` levels from there
            fields = self.get_field(prefix).available_children
        else:
            prefix = []
            fields = self.root_fields()

        output = {}
        for path, field in fields.items():

            output[".".join(prefix + [path])] = field
            for subpath, child in _children_at_depth(field, depth - 1):
                output[".".join(prefix + [path] + subpath)] = child
        return output

    def root_fields(self):  # pragma: no cover
        """Return the root fields for this starting object.

        Must be overridden in a subclass actually implementing
        a starting object.
        """
        raise NotImplementedError("Must be implemented in BaseStartingObject subclass")

    def get_field(self, path):
        """Return a single field with the given path."""
        children = self.root_fields()
        for elem in path:
            field = children[elem]
            children = field.available_children
        return field


class CaseStartingObject(BaseStartingObject):
    identifier = "cases"
    label = "Cases"

    def get_query(self):
        return self.visibility_source.cases()

    def root_fields(self):
        direct_fields = {
            "meta": MetaField(
                parent=None,
                identifier="meta",
                label="Meta",
                visibility_source=self.visibility_source,
            ),
            "closed_at": AttributeField(
                parent=None,
                identifier="closed_at",
                is_date=True,
                visibility_source=self.visibility_source,
            ),
            "created_at": AttributeField(
                parent=None,
                identifier="created_at",
                is_date=True,
                visibility_source=self.visibility_source,
            ),
            "status": AttributeField(
                parent=None,
                identifier="status",
                visibility_source=self.visibility_source,
            ),
            "id": AttributeField(
                parent=None,
                identifier="id",
                visibility_source=self.visibility_source,
            ),
            "document[*]": FormDocumentField(
                identifier="document[*]",
                parent=None,
                visibility_source=self.visibility_source,
            ),
        }

        case_forms = form_models.Form.objects.filter(
            documents__case__isnull=False
        ).distinct()

        form_fields = {
            f"document[{form.slug}]": FormDocumentField(
                identifier=f"document[{form.slug}]",
                parent=None,
                visibility_source=self.visibility_source,
            )
            for form in case_forms
        }

        tasks = workflow_models.Task.objects.all()

        workitem_fields = {
            f"workitem[{task.slug},{selector}]": WorkItemField(
                identifier=f"workitem[{task.slug},{selector}]",
                parent=None,
                visibility_source=self.visibility_source,
            )
            for selector in ["first", "last", "firstclosed", "lastclosed"]
            for task in tasks
        }

        return {**direct_fields, **form_fields, **workitem_fields}


class WorkItemsStartingObject(BaseStartingObject):
    identifier = "work_items"
    label = "Work Items"

    def get_query(self):
        return self.visibility_source.work_items()

    def root_fields(self):
        direct_fields = {
            "meta": MetaField(
                parent=None,
                identifier="meta",
                label="Meta",
                visibility_source=self.visibility_source,
            ),
            "task_id": AttributeField(
                parent=None,
                identifier="task_id",
                is_date=False,
                visibility_source=self.visibility_source,
            ),
            "closed_at": AttributeField(
                parent=None,
                identifier="closed_at",
                is_date=True,
                visibility_source=self.visibility_source,
            ),
            "created_at": AttributeField(
                parent=None,
                identifier="created_at",
                is_date=True,
                visibility_source=self.visibility_source,
            ),
            "status": AttributeField(
                parent=None,
                identifier="status",
                visibility_source=self.visibility_source,
            ),
            "id": AttributeField(
                parent=None,
                identifier="id",
                visibility_source=self.visibility_source,
            ),
        }

        workitem_forms = form_models.Form.objects.filter(
            documents__work_item__isnull=False
        ).distinct()

        form_fields = {
            f"document[{form.slug}]": FormDocumentField(
                identifier=f"document[{form.slug}]",
                parent=None,
                visibility_source=self.visibility_source,
            )
            for form in workitem_forms
        }

        return {**direct_fields, **form_fields}


class DocumentsStartingObject(BaseStartingObject):
    identifier = "documents"
    label = "documents"

    def get_query(self):
        return self.visibility_source.documents()

    def root_fields(self):
        direct_fields = {
            "id": AttributeField(
                parent=None,
                identifier="id",
                visibility_source=self.visibility_source,
            ),
            "meta": MetaField(
                parent=None,
                identifier="meta",
                label="Meta",
                visibility_source=self.visibility_source,
            ),
            "created_at": AttributeField(
                parent=None,
                identifier="created_at",
                is_date=True,
                visibility_source=self.visibility_source,
            ),
            "caluma_form": FormInfoField(
                parent=None,
                identifier="caluma_form",
                visibility_source=self.visibility_source,
            ),
        }

        form_fields = {
            f"answers[{form.slug}]": DirectDocumentField(
                identifier=f"answers[{form.slug}]",
                parent=None,
                visibility_source=self.visibility_source,
            )
            for form in form_models.Form.objects.all()
        }

        return {**direct_fields, **form_fields}


class SimpleTable:
    class _AnonymousInfo:
        """
        Pseudo GraphQL info object.

        Used for commandline access, so the commandline version
        can also invoke the visibilities.
        """

        context = {}

    def __init__(self, table, info=_AnonymousInfo):
        self.info = info
        self.table = table
        self.last_query = None
        self.last_query_params = None

        self.starting_object = BaseStartingObject.get_object(
            self.table.starting_object,
            self.info,
            disable_visibilities=self.table.disable_visibilities,
        )

        self.base_query = sql.Query(from_=self.starting_object.get_query())

    @cached_property
    def _fields(self):
        field_spec_qs = self.table.fields.all()
        # sort fields to join by length. This way, parent fields
        # will be already joined when their children get added, so
        # they will have the right alias.
        table_fields = sorted(field_spec_qs, key=lambda f: len(f.data_source))

        fields = {}
        for field_spec in table_fields:
            fields[field_spec.alias] = self.starting_object.get_field(
                field_spec.data_source.split(".")
            )
            fields[field_spec.alias].alias = field_spec.alias
            fields[field_spec.alias].show_output = field_spec.show_output

            # Filters must always be on the outermost (base) query,
            # as we're using LEFT JOINs to get at related data.
            self.base_query.add_field_filter(
                self._sql_alias(field_spec.alias), field_spec.filters
            )
        return fields

    @cached_property
    def field_ordering(self):
        return list(
            self.table.fields.all()
            .filter(show_output=True)
            .values_list("alias", flat=True)
        )

    def get_sql_and_params(self):
        """Return a list of records as specified in the given table config."""

        base_query = self.get_query_object()

        sql_query, params, _ = sql.QueryRender(base_query).as_sql(alias=None)

        self.last_query = sql_query
        self.last_query_params = params
        return sql_query, params

    def get_query_object(self):

        fields = self._fields

        step_queries = {}
        base_query = deepcopy(self.base_query)

        for alias, field in fields.items():
            query = base_query

            field.required = True
            sql_field = field.query_field()
            sql_field.alias = self._sql_alias(alias)

            cache_path = ""
            for step in sql_field.path_from_root():
                cache_path = f"{cache_path}.{step.identifier}"
                if cache_path in step_queries:
                    query = step_queries[cache_path]
                else:
                    query = step.annotate(query)
                    step_queries[cache_path] = query

        return base_query

    def get_records(self):

        sql_query, params = self.get_sql_and_params()

        with connection.connection.cursor(cursor_factory=DictCursor) as cursor:
            cursor.execute(sql_query, params)
            data = cursor.fetchall()

            return [
                {
                    field.alias: field.parse_value(
                        row[f"analytics_result_{field.alias}"]
                    )
                    for field in self._fields.values()
                    if field.show_output
                }
                for row in data
            ]

    def get_summary(self):
        # simple tables can't do summary, but we must implement
        # it as both table types share the same GQL interface
        return {}  # pragma: no cover

    def _sql_alias(self, user_alias):
        return f"analytics_result_{user_alias}"



import graphene
from django_filters.rest_framework import BooleanFilter, MultipleChoiceFilter

from ..caluma_core.filters import (
    BaseFilterSet,
    CharFilter,
    GlobalIDFilter,
    GlobalIDMultipleChoiceFilter,
    JSONValueFilter,
    MetaFilterSet,
    SearchFilter,
    StringListFilter,
    generate_list_filter_class,
)
from ..caluma_core.ordering import AttributeOrderingFactory, MetaFieldOrdering
from ..caluma_form.filters import HasAnswerFilter, SearchAnswersFilter
from ..caluma_form.ordering import AnswerValueOrdering
from . import models


def case_status_filter(*args, **kwargs):
    case_status_descriptions = {
        s.upper(): d for s, d in models.Case.STATUS_CHOICE_TUPLE
    }

    class EnumWithDescriptionsType(object):
        @property
        def description(self):
            return case_status_descriptions[self.name]

    enum = graphene.Enum(
        "CaseStatusArgument",
        [(key.upper(), key) for (key, desc) in models.Case.STATUS_CHOICE_TUPLE],
        type=EnumWithDescriptionsType,
    )
    return generate_list_filter_class(enum)(*args, **kwargs)


class WorkflowFilterSet(MetaFilterSet):
    search = SearchFilter(fields=("slug", "name", "description"))

    class Meta:
        model = models.Workflow
        fields = ("slug", "name", "description", "is_published", "is_archived")


class WorkflowOrderSet(BaseFilterSet):
    meta = MetaFieldOrdering()
    attribute = AttributeOrderingFactory(
        models.Workflow,
        fields=[
            "created_at",
            "modified_at",
            "allow_all_forms",
            "description",
            "is_archived",
            "is_published",
            "name",
            "slug",
        ],
    )

    class Meta:
        model = models.Workflow
        fields = ("meta", "attribute")


class FlowFilterSet(BaseFilterSet):
    task = GlobalIDFilter(field_name="task_flows__task")

    class Meta:
        model = models.Flow
        fields = ("task",)


class FlowOrderSet(BaseFilterSet):
    meta = MetaFieldOrdering()
    attribute = AttributeOrderingFactory(
        models.Flow,
        fields=["created_at", "modified_at", "task"],
    )

    class Meta:
        model = models.Flow
        fields = ("meta", "attribute")


class CaseFilterSet(MetaFilterSet):
    id = GlobalIDFilter()
    id.deprecation_reason = "Use ids filter instead"
    ids = GlobalIDMultipleChoiceFilter(field_name="pk")

    document_form = CharFilter(field_name="document__form_id")
    document_forms = MultipleChoiceFilter(field_name="document__form_id")
    has_answer = HasAnswerFilter(document_id="document__pk")
    work_item_document_has_answer = HasAnswerFilter(
        document_id="work_items__document__pk"
    )
    root_case = GlobalIDFilter(field_name="family")
    search_answers = SearchAnswersFilter(document_id="document__pk")
    status = case_status_filter(lookup_expr="in")
    exclude_child_cases = BooleanFilter(method="filter_exclude_child_cases")

    def filter_exclude_child_cases(self, queryset, name, value):
        if not value:
            return queryset

        return queryset.filter(parent_work_item__isnull=True)

    class Meta:
        model = models.Case
        fields = ("workflow",)


class CaseOrderSet(BaseFilterSet):
    meta = MetaFieldOrdering()
    attribute = AttributeOrderingFactory(
        models.Case,
        fields=[
            "created_at",
            "modified_at",
            "status",
            "document__form__name",
        ],
    )
    document_answer = AnswerValueOrdering(document_via="document")

    class Meta:
        model = models.Case
        fields = ("meta", "attribute", "document_answer")


class TaskFilterSet(MetaFilterSet):
    search = SearchFilter(fields=("slug", "name", "description"))

    class Meta:
        model = models.Task
        fields = ("slug", "name", "description", "type", "is_archived")


class TaskOrderSet(BaseFilterSet):
    meta = MetaFieldOrdering()
    attribute = AttributeOrderingFactory(
        models.Task,
        fields=[
            "created_at",
            "modified_at",
            "lead_time",
            "type",
            "description",
            "is_archived",
            "name",
            "slug",
        ],
    )

    class Meta:
        model = models.Task
        fields = ("meta", "attribute")


class WorkItemFilterSet(MetaFilterSet):
    id = GlobalIDFilter()
    addressed_groups = StringListFilter(lookup_expr="overlap")
    controlling_groups = StringListFilter(lookup_expr="overlap")
    assigned_users = StringListFilter(lookup_expr="overlap")

    document_has_answer = HasAnswerFilter(document_id="document__pk")
    case_document_has_answer = HasAnswerFilter(document_id="case__document__pk")
    case_meta_value = JSONValueFilter(field_name="case__meta")
    root_case_meta_value = JSONValueFilter(field_name="case__family__meta")
    case_search_answers = SearchAnswersFilter(document_id="case__document__pk")

    tasks = MultipleChoiceFilter(field_name="task_id")

    has_deadline = BooleanFilter(
        field_name="deadline", lookup_expr="isnull", exclude=True
    )
    case_family = GlobalIDFilter(field_name="case__family")

    class Meta:
        model = models.WorkItem
        fields = (
            "status",
            "name",
            "task",
            "tasks",
            "case",
            "created_at",
            "closed_at",
            "modified_at",
            "deadline",
            "has_deadline",
            "case_family",
        )


class WorkItemOrderSet(BaseFilterSet):
    meta = MetaFieldOrdering()
    case_meta = MetaFieldOrdering(field_name="case__meta")
    attribute = AttributeOrderingFactory(
        models.WorkItem,
        fields=[
            "created_at",
            "modified_at",
            "closed_at",
            "description",
            "name",
            "deadline",
            "status",
            "slug",
            "case__document__form__name",
        ],
    )
    document_answer = AnswerValueOrdering(document_via="document")
    case_document_answer = AnswerValueOrdering(document_via="case__document")

    class Meta:
        model = models.WorkItem
        fields = (
            "meta",
            "case_meta",
            "attribute",
            "document_answer",
            "case_document_answer",
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.17 on 2018-12-12 12:43
from __future__ import unicode_literals

import uuid

import django.db.models.deletion
import localized_fields.fields.field
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [("caluma_form", "0001_initial")]

    operations = [
        migrations.CreateModel(
            name="Case",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            (
                                "running",
                                "Case is running and work items need to be completed.",
                            ),
                            ("completed", "Case is done."),
                            ("canceled", "Case is cancelled."),
                        ],
                        db_index=True,
                        max_length=50,
                    ),
                ),
                ("meta", models.JSONField(default={})),
                (
                    "document",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="cases",
                        to="caluma_form.Document",
                    ),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="Flow",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("next", models.TextField()),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="Task",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                ("slug", models.SlugField(primary_key=True, serialize=False)),
                ("name", localized_fields.fields.field.LocalizedField(required=[])),
                (
                    "description",
                    localized_fields.fields.field.LocalizedField(
                        blank=True, null=True, required=[]
                    ),
                ),
                (
                    "type",
                    models.CharField(
                        choices=[
                            ("simple", "Task which can only be marked as completed.")
                        ],
                        max_length=50,
                    ),
                ),
                ("meta", models.JSONField(default={})),
                ("is_archived", models.BooleanField(default=False)),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="TaskFlow",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "flow",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="task_flows",
                        to="caluma_workflow.Flow",
                    ),
                ),
                (
                    "task",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="task_flows",
                        to="caluma_workflow.Task",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Workflow",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                ("slug", models.SlugField(primary_key=True, serialize=False)),
                ("name", localized_fields.fields.field.LocalizedField(required=[])),
                (
                    "description",
                    localized_fields.fields.field.LocalizedField(
                        blank=True, null=True, required=[]
                    ),
                ),
                ("meta", models.JSONField(default={})),
                ("is_published", models.BooleanField(default=False)),
                ("is_archived", models.BooleanField(default=False)),
                (
                    "form",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="workflows",
                        to="caluma_form.Form",
                    ),
                ),
                (
                    "start",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="caluma_workflow.Task",
                    ),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="WorkItem",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("ready", "Task is ready to be processed."),
                            ("completed", "Task is done."),
                            ("canceled", "Task is cancelled."),
                        ],
                        db_index=True,
                        max_length=50,
                    ),
                ),
                ("meta", models.JSONField(default={})),
                (
                    "case",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="work_items",
                        to="caluma_workflow.Case",
                    ),
                ),
                (
                    "child_case",
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="parent_work_item",
                        to="caluma_workflow.Case",
                    ),
                ),
                (
                    "task",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="work_items",
                        to="caluma_workflow.Task",
                    ),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.AddField(
            model_name="taskflow",
            name="workflow",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="task_flows",
                to="caluma_workflow.Workflow",
            ),
        ),
        migrations.AddField(
            model_name="case",
            name="workflow",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="cases",
                to="caluma_workflow.Workflow",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="taskflow", unique_together=set([("workflow", "task")])
        ),
    ]



from graphene import ConnectionField, JSONString, String
from graphene.types import ObjectType

from ..caluma_core.types import CountableConnectionBase
from ..caluma_form.models import Question
from .data_source_handlers import get_data_source_data, get_data_sources


class DataSource(ObjectType):
    info = String()
    name = String(required=True)


class DataSourceConnection(CountableConnectionBase):
    class Meta:
        node = DataSource


class DataSourceData(ObjectType):
    label = String(required=True)
    slug = String(required=True)


class DataSourceDataConnection(CountableConnectionBase):
    class Meta:
        node = DataSourceData


class Query(ObjectType):
    all_data_sources = ConnectionField(DataSourceConnection)
    data_source = ConnectionField(
        DataSourceDataConnection,
        name=String(required=True),
        question=String(
            description="Slug of the question passed as context to the data source"
        ),
        context=JSONString(
            description="JSON object passed as context to the data source"
        ),
    )

    def resolve_all_data_sources(self, info, **kwargs):
        return get_data_sources()

    def resolve_data_source(self, info, name, question=None, context=None, **kwargs):
        if question:
            question = Question.objects.get(pk=question)

        return get_data_source_data(info.context.user, name, question, context)
